import asyncio
import platform
import threading
from typing import TYPE_CHECKING

from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class StopCommand:
    def stop(self,  # type: HummingbotApplication
             skip_order_cancellation: bool = False):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.stop, skip_order_cancellation)
            return
        safe_ensure_future(self.stop_loop(skip_order_cancellation), loop=self.ev_loop)

    async def stop_loop(self,  # type: HummingbotApplication
                        skip_order_cancellation: bool = False):
        self.logger().info("stop command initiated.")
        self.notify("\nWinding down...")

        # Restore App Nap on macOS.
        if platform.system() == "Darwin":
            import appnope
            appnope.nap()

        if isinstance(self.strategy, ScriptStrategyBase):
            await self.strategy.on_stop()

        if self._trading_required and not skip_order_cancellation:
            # Remove the strategy from clock before cancelling orders, to
            # prevent race condition where the strategy tries to create more
            # orders during cancellation.
            if self.clock:
                self.clock.remove_iterator(self.strategy)
            success = await self._cancel_outstanding_orders()
            # Give some time for cancellation events to trigger
            await asyncio.sleep(2)
            if success:
                # Only erase markets when cancellation has been successful
                self.markets = {}

        if self.strategy_task is not None and not self.strategy_task.cancelled():
            self.strategy_task.cancel()

        if RateOracle.get_instance().started:
            RateOracle.get_instance().stop()

        if self.markets_recorder is not None:
            self.markets_recorder.stop()

        if self.kill_switch is not None:
            self.kill_switch.stop()

        self.strategy_task = None
        self.strategy = None
        self.market_pair = None
        self.clock = None
        self.markets_recorder = None
        self.market_trading_pairs_map.clear()



import gzip
import json
import os
import platform
from collections import namedtuple
from hashlib import md5
from typing import Any, Callable, Dict, Optional, Tuple

from hexbytes import HexBytes

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.api_throttler.async_throttler_base import AsyncThrottlerBase
from hummingbot.core.utils.tracking_nonce import NonceCreator, get_tracking_nonce
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSResponse
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_post_processors import WSPostProcessorBase

TradeFillOrderDetails = namedtuple("TradeFillOrderDetails", "market exchange_trade_id symbol")


def build_api_factory(throttler: AsyncThrottlerBase) -> WebAssistantsFactory:
    throttler = throttler or AsyncThrottler(rate_limits=[])
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def split_hb_trading_pair(trading_pair: str) -> Tuple[str, str]:
    base, quote = trading_pair.split("-")
    return base, quote


def combine_to_hb_trading_pair(base: str, quote: str) -> str:
    trading_pair = f"{base}-{quote}"
    return trading_pair


def validate_trading_pair(trading_pair: str) -> bool:
    valid = False
    if "-" in trading_pair and len(trading_pair.split("-")) == 2:
        valid = True
    return valid


def _bot_instance_id() -> str:
    return md5(f"{platform.uname()}_pid:{os.getpid()}_ppid:{os.getppid()}".encode("utf-8")).hexdigest()


def get_new_client_order_id(
    is_buy: bool, trading_pair: str, hbot_order_id_prefix: str = "", max_id_len: Optional[int] = None
) -> str:
    """
    Creates a client order id for a new order

    Note: If the need for much shorter IDs arises, an option is to concatenate the host name, the PID,
    and the nonce, and hash the result.

    :param is_buy: True if the order is a buy order, False otherwise
    :param trading_pair: the trading pair the order will be operating with
    :param hbot_order_id_prefix: The hummingbot-specific identifier for the given exchange
    :param max_id_len: The maximum length of the ID string.
    :return: an identifier for the new order to be used in the client
    """
    side = "B" if is_buy else "S"
    symbols = split_hb_trading_pair(trading_pair)
    base = symbols[0].upper()
    quote = symbols[1].upper()
    base_str = f"{base[0]}{base[-1]}"
    quote_str = f"{quote[0]}{quote[-1]}"
    client_instance_id = _bot_instance_id()
    ts_hex = hex(get_tracking_nonce())[2:]
    client_order_id = f"{hbot_order_id_prefix}{side}{base_str}{quote_str}{ts_hex}{client_instance_id}"

    if max_id_len is not None:
        id_prefix = f"{hbot_order_id_prefix}{side}{base_str}{quote_str}"
        suffix_max_length = max_id_len - len(id_prefix)
        if suffix_max_length < len(ts_hex):
            id_suffix = md5(f"{ts_hex}{client_instance_id}".encode()).hexdigest()
            client_order_id = f"{id_prefix}{id_suffix[:suffix_max_length]}"
        else:
            client_order_id = client_order_id[:max_id_len]
    return client_order_id


def get_new_numeric_client_order_id(nonce_creator: NonceCreator, max_id_bit_count: Optional[int] = None) -> int:
    hexa_hash = _bot_instance_id()
    host_part = int(hexa_hash, 16)
    client_order_id = int(f"{host_part}{nonce_creator.get_tracking_nonce()}")
    if max_id_bit_count:
        max_int = 2 ** max_id_bit_count - 1
        client_order_id &= max_int
    return client_order_id


class TimeSynchronizerRESTPreProcessor(RESTPreProcessorBase):
    """
    This pre processor is intended to be used in those connectors that require synchronization with the server time
    to accept API requests. It ensures the synchronizer has at least one server time sample before being used.
    """

    def __init__(self, synchronizer: TimeSynchronizer, time_provider: Callable):
        super().__init__()
        self._synchronizer = synchronizer
        self._time_provider = time_provider

    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        await self._synchronizer.update_server_time_if_not_initialized(time_provider=self._time_provider())
        return request


class GZipCompressionWSPostProcessor(WSPostProcessorBase):
    """
    Performs the necessary response processing from both public and private websocket streams.
    """

    async def post_process(self, response: WSResponse) -> WSResponse:
        if not isinstance(response.data, bytes):
            # Unlike Market WebSocket, the return data of Account and Order Websocket are not compressed by GZIP.
            return response
        encoded_msg: bytes = gzip.decompress(response.data)
        msg: Dict[str, Any] = json.loads(encoded_msg.decode("utf-8"))

        return WSResponse(data=msg)


def to_0x_hex(signature: HexBytes | bytes) -> str:
    """
    Convert a string to a 0x-prefixed hex string.
    """
    if hasattr(signature, "to_0x_hex"):
        return signature.to_0x_hex()

    return hex if (hex := signature.hex()).startswith("0x") else f"0x{hex}"



from hummingbot.core.api_throttler.data_types import RateLimit

EXCHANGE_NAME = "bitmart_perpetual"
BROKER_ID = "hummingbotfound"
MAX_ORDER_ID_LEN = 32

DOMAIN = EXCHANGE_NAME

PERPETUAL_BASE_URL = "https://api-cloud-v2.bitmart.com"
PERPETUAL_WS_URL = "wss://openapi-ws-v2.bitmart.com"

SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 20
PUBLIC_WS_ENDPOINT = "/api?protocol=1.1"
PRIVATE_WS_ENDPOINT = "/user?protocol=1.1"

TIME_IN_FORCE_GTC = 1  # Good till cancelled
TIME_IN_FORCE_FOK = 2  # Fill or kill
TIME_IN_FORCE_IOC = 3  # Immediate or cancel
TIME_IN_FORCE_MAKER_ONLY = 4  # Good Till Crossing

# Public API v2 Endpoints
SNAPSHOT_REST_URL = "/contract/public/depth"  # 12 times / 2 sec
EXCHANGE_INFO_URL = "/contract/public/details"  # 12 times / 2 sec
FUNDING_INFO_URL = "/contract/public/funding-rate"  # 12 times / 2 sec
SERVER_TIME_PATH_URL = "/system/time"  # 10 times / 2 sec

# Private API v1 Endpoints
SUBMIT_ORDER_URL = "/contract/private/submit-order"  # 24 times / 2 sec
ORDER_DETAILS = "/contract/private/order"  # 50 times / 2 sec
CANCEL_ORDER_URL = "/contract/private/cancel-order"  # 40 times / 2 sec
ACCOUNT_TRADE_LIST_URL = "/contract/private/trades"  # 6 times / 2 sec
SET_LEVERAGE_URL = "/contract/private/submit-leverage"  # 24 times / 2 sec
GET_INCOME_HISTORY_URL = "/contract/private/transaction-history"  # 6 times / 2 sec

# Private API v2 Endpoints
ASSETS_DETAIL = "/contract/private/assets-detail"  # 12 times / 2 sec
POSITION_INFORMATION_URL = "/contract/private/position"  # 6 times / 2 sec

# Public WS channels
TRADE_STREAM_CHANNEL = "futures/trade"
FUNDING_INFO_CHANNEL = "futures/fundingRate"
TICKERS_CHANNEL = "futures/ticker"
ORDER_BOOK_CHANNEL = "futures/depthIncrease50"
ORDER_BOOK_SPEED = "200ms"

# Private WS channels
WS_POSITIONS_CHANNEL = "futures/position"
WS_ORDERS_CHANNEL = "futures/order"
WS_ACCOUNT_CHANNEL = "futures/asset:USDT"

# Funding Settlement Time Span
FUNDING_SETTLEMENT_DURATION = (0, 30)  # seconds before snapshot, seconds after snapshot

# Rate Limit Type
REQUEST_WEIGHT = "REQUEST_WEIGHT"
ORDERS_1MIN = "ORDERS_1MIN"
ORDERS_1SEC = "ORDERS_1SEC"

HEARTBEAT_TIME_INTERVAL = 30.0

# Rate Limit time intervals
RATE_LIMITS = [
    # Weight Limits for individual endpoints
    RateLimit(limit_id=SNAPSHOT_REST_URL, limit=12, time_interval=2),
    RateLimit(limit_id=EXCHANGE_INFO_URL, limit=12, time_interval=2),
    RateLimit(limit_id=FUNDING_INFO_URL, limit=12, time_interval=2),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=10, time_interval=2),
    RateLimit(limit_id=SUBMIT_ORDER_URL, limit=24, time_interval=2),
    RateLimit(limit_id=ORDER_DETAILS, limit=50, time_interval=2),
    RateLimit(limit_id=CANCEL_ORDER_URL, limit=40, time_interval=2),
    RateLimit(limit_id=ACCOUNT_TRADE_LIST_URL, limit=6, time_interval=2),
    RateLimit(limit_id=SET_LEVERAGE_URL, limit=24, time_interval=2),
    RateLimit(limit_id=GET_INCOME_HISTORY_URL, limit=6, time_interval=2),
    RateLimit(limit_id=ASSETS_DETAIL, limit=12, time_interval=2),
    RateLimit(limit_id=POSITION_INFORMATION_URL, limit=6, time_interval=2),
]

CODE_OK = 1000
ORDER_NOT_EXIST_ERROR_CODE = 40035
ORDER_NOT_EXIST_MESSAGE = "The order is not exist"
UNKNOWN_ORDER_ERROR_CODE = 40037
UNKNOWN_ORDER_MESSAGE = "The order id is not exist"



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS
from hummingbot.connector.utilities.oms_connector.oms_connector_auth import OMSConnectorAuth
from hummingbot.connector.utilities.oms_connector.oms_connector_web_utils import (
    OMSConnectorURLCreatorBase,
    OMSConnectorWebAssistantsFactory,
)
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.rest_assistant import RESTAssistant
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.utilities.oms_connector.oms_connector_exchange import OMSExchange


class OMSConnectorAPIOrderBookDataSource(OrderBookTrackerDataSource):
    def __init__(
        self,
        trading_pairs: List[str],
        connector: 'OMSExchange',
        api_factory: OMSConnectorWebAssistantsFactory,
        url_provider: OMSConnectorURLCreatorBase,
        oms_id: int,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._rest_assistant: Optional[RESTAssistant] = None
        self._ws_assistant: Optional[WSAssistant] = None
        self._auth: OMSConnectorAuth = api_factory.auth
        self._url_provider = url_provider
        self._oms_id = oms_id
        self._nonce_provider = NonceCreator.for_milliseconds()

    async def get_last_traded_prices(
        self, trading_pairs: List[str], domain: Optional[str] = None
    ) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _parse_trade_message(self, raw_message: List[Dict[int, Union[int, float]]], message_queue: asyncio.Queue):
        raise NotImplementedError  # OMS connectors do not provide a public trades endpoint

    async def _parse_order_book_diff_message(
        self, raw_message: List[List[Union[int, float]]], message_queue: asyncio.Queue
    ):
        msg_data = raw_message[CONSTANTS.MSG_DATA_FIELD]
        first_row = msg_data[0]
        ts_ms = first_row[CONSTANTS.DIFF_UPDATE_TS_FIELD]
        update_id = self._nonce_provider.get_tracking_nonce(timestamp=ts_ms * 1e-3)
        instrument_id = first_row[CONSTANTS.DIFF_UPDATE_INSTRUMENT_ID_FIELD]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=str(instrument_id))
        bids, asks = self._get_bids_and_asks_from_snapshot(msg_data)

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": bids,
            "asks": asks,
        }
        diff_message: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.DIFF,
            order_book_message_content,
            ts_ms * 1e-3,
        )

        message_queue.put_nowait(diff_message)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response = await self._request_order_book_snapshot(trading_pair)
        first_row = snapshot_response[0]
        ts_ms = first_row[CONSTANTS.DIFF_UPDATE_TS_FIELD]
        update_id = self._nonce_provider.get_tracking_nonce(timestamp=ts_ms * 1e-3)
        bids, asks = self._get_bids_and_asks_from_snapshot(snapshot_response)

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": bids,
            "asks": asks,
        }
        snapshot_msg = OrderBookMessage(OrderBookMessageType.SNAPSHOT, order_book_message_content, ts_ms)
        return snapshot_msg

    @staticmethod
    def _get_bids_and_asks_from_snapshot(
        snapshot: List[List[Union[int, float]]]
    ) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:
        """OMS connectors do not guarantee that the data is sorted in any way."""
        asks = []
        bids = []
        for row in snapshot:
            update = (row[CONSTANTS.DIFF_UPDATE_PRICE_FIELD], row[CONSTANTS.DIFF_UPDATE_AMOUNT_FIELD])
            if row[CONSTANTS.DIFF_UPDATE_SIDE_FIELD] == CONSTANTS.BUY_ACTION:
                bids.append(update)
            else:
                asks.append(update)
        return bids, asks

    async def _request_order_book_snapshot(self, trading_pair: str) -> List[List[Union[int, float]]]:
        instrument_id = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params = {
            CONSTANTS.OMS_ID_FIELD: self._oms_id,
            CONSTANTS.INSTRUMENT_ID_FIELD: int(instrument_id),
            CONSTANTS.DEPTH_FIELD: CONSTANTS.MAX_L2_SNAPSHOT_DEPTH,
        }
        rest_assistant = await self._get_rest_assistant()
        url = self._url_provider.get_rest_url(path_url=CONSTANTS.REST_GET_L2_SNAPSHOT_ENDPOINT)
        resp = await rest_assistant.execute_request(
            url=url,
            throttler_limit_id=CONSTANTS.REST_REQ_LIMIT_ID,
            params=params,
            method=RESTMethod.GET,
        )
        return resp

    async def _get_rest_assistant(self) -> RESTAssistant:
        if self._rest_assistant is None:
            self._ensure_authenticated()
            self._rest_assistant = await self._api_factory.get_rest_assistant()
        return self._rest_assistant

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._get_ws_assistant()
        url = self._url_provider.get_ws_url()
        await ws.connect(ws_url=url, message_timeout=CONSTANTS.WS_MESSAGE_TIMEOUT)
        return ws

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ensure_authenticated()
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    def _ensure_authenticated(self):
        if not self._auth.initialized:
            raise RuntimeError("The authenticator is not initialized.")

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            for trading_pair in self._trading_pairs:
                instrument_id = await self._connector.exchange_symbol_associated_to_pair(trading_pair)
                req_params = {
                    CONSTANTS.OMS_ID_FIELD: self._oms_id,
                    CONSTANTS.INSTRUMENT_ID_FIELD: int(instrument_id),
                    CONSTANTS.DEPTH_FIELD: CONSTANTS.MAX_L2_SNAPSHOT_DEPTH,
                }
                payload = {
                    CONSTANTS.MSG_ENDPOINT_FIELD: CONSTANTS.WS_L2_SUB_ENDPOINT,
                    CONSTANTS.MSG_DATA_FIELD: req_params,
                }
                subscribe_orderbook_request = WSJSONRequest(payload=payload)

                async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_REQ_LIMIT_ID):
                    await ws.send(subscribe_orderbook_request)

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if event_message[CONSTANTS.MSG_TYPE_FIELD] != CONSTANTS.ERROR_MSG_TYPE:
            event_channel = event_message[CONSTANTS.MSG_ENDPOINT_FIELD]
            if event_channel == CONSTANTS.WS_L2_EVENT:
                channel = self._diff_messages_queue_key
        return channel

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        while True:
            try:
                async for ws_response in websocket_assistant.iter_messages():
                    data: Dict[str, Any] = ws_response.data
                    channel: str = self._channel_originating_message(event_message=data)
                    if channel in [self._diff_messages_queue_key, self._trade_messages_queue_key]:
                        self._message_queue[channel].put_nowait(data)
            except asyncio.TimeoutError:
                ping_payload = {
                    CONSTANTS.MSG_ENDPOINT_FIELD: CONSTANTS.WS_PING_REQUEST,
                    CONSTANTS.MSG_DATA_FIELD: {},
                }
                ping_request = WSJSONRequest(payload=ping_payload)
                async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_PING_REQUEST):
                    await websocket_assistant.send(request=ping_request)



import logging
from collections import defaultdict
from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.data_type.common import PositionAction, TradeType
from hummingbot.core.data_type.trade_fee import DeductedFromReturnsTradeFee, TokenAmount
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.logger import HummingbotLogger
from hummingbot.model.trade_fill import TradeFill

s_decimal_0 = Decimal("0")
s_decimal_nan = Decimal("NaN")


@dataclass
class PerformanceMetrics:
    _logger = None

    num_buys: int = 0
    num_sells: int = 0
    num_trades: int = 0

    b_vol_base: Decimal = s_decimal_0
    s_vol_base: Decimal = s_decimal_0
    tot_vol_base: Decimal = s_decimal_0

    b_vol_quote: Decimal = s_decimal_0
    s_vol_quote: Decimal = s_decimal_0
    tot_vol_quote: Decimal = s_decimal_0

    avg_b_price: Decimal = s_decimal_0
    avg_s_price: Decimal = s_decimal_0
    avg_tot_price: Decimal = s_decimal_0

    start_base_bal: Decimal = s_decimal_0
    start_quote_bal: Decimal = s_decimal_0
    cur_base_bal: Decimal = s_decimal_0
    cur_quote_bal: Decimal = s_decimal_0
    start_price: Decimal = s_decimal_0
    cur_price: Decimal = s_decimal_0
    start_base_ratio_pct: Decimal = s_decimal_0
    cur_base_ratio_pct: Decimal = s_decimal_0

    hold_value: Decimal = s_decimal_0
    cur_value: Decimal = s_decimal_0
    trade_pnl: Decimal = s_decimal_0

    fee_in_quote: Decimal = s_decimal_0
    total_pnl: Decimal = s_decimal_0
    return_pct: Decimal = s_decimal_0

    def __init__(self):
        # fees is a dictionary of token and total fee amount paid in that token.
        self.fees: Dict[str, Decimal] = defaultdict(lambda: s_decimal_0)

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    @classmethod
    async def create(cls,
                     trading_pair: str,
                     trades: List[Any],
                     current_balances: Dict[str, Decimal]) -> 'PerformanceMetrics':
        performance = PerformanceMetrics()
        await performance._initialize_metrics(trading_pair, trades, current_balances)
        return performance

    @staticmethod
    def position_order(open: list, close: list) -> Tuple[Any, Any]:
        """
        Pair open position order with close position orders
        :param open: a list of orders that may have an open position order
        :param close: a list of orders that may have an close position order
        :return: A tuple containing a pair of an open order with a close position order
        """
        result = None

        try:
            first_with_open_position = next((order for order in open if order.position == "OPEN"))
            first_with_close_position = next((order for order in close if order.position == "CLOSE"))
            open.remove(first_with_open_position)
            close.remove(first_with_close_position)
            result = (first_with_open_position, first_with_close_position)
        except StopIteration:
            pass

        return result

    @staticmethod
    def aggregate_orders(orders: list) -> list:
        grouped_orders = {}
        for order in orders:
            group = grouped_orders.get(order.order_id, [])
            group.append(order)
            grouped_orders[order.order_id] = group

        aggregated_orders = []
        for group in grouped_orders.values():
            aggregated_prices = 0
            aggregated_amounts = 0
            for order in group:
                aggregated_prices += order.price
                aggregated_amounts += order.amount
            aggregated = group[0]
            aggregated.price = aggregated_prices / len(group)
            aggregated.amount = aggregated_amounts
            aggregated_orders.append(aggregated)

        return aggregated_orders

    @staticmethod
    def aggregate_position_order(buys: list, sells: list) -> Tuple[list, list]:
        """
        Aggregate the amount field for orders with multiple fills
        :param buys: a list of buy orders
        :param sells: a list of sell orders
        :return: 2 lists containing aggregated amounts for buy and sell orders.
        """
        aggregated_buys = PerformanceMetrics.aggregate_orders(buys)
        aggregated_sells = PerformanceMetrics.aggregate_orders(sells)

        return aggregated_buys, aggregated_sells

    @staticmethod
    def derivative_pnl(long: list, short: list) -> List[Decimal]:
        # It is assumed that the amount and leverage for both open and close orders are the same.
        """
        Calculates PnL for a close position
        :param long: a list containing pairs of open and closed long position orders
        :param short: a list containing pairs of open and closed short position orders
        :return: A list containing PnL for each closed positions
        """
        pnls = []
        for lg in long:
            pnls.append((lg[1].price - lg[0].price) * lg[1].amount)
        for st in short:
            pnls.append((st[0].price - st[1].price) * st[1].amount)
        return pnls

    @staticmethod
    def smart_round(value: Decimal, precision: Optional[int] = None) -> Decimal:
        if value is None or value.is_nan():
            return value
        if precision is not None:
            precision = 1 / (10 ** precision)
            return Decimal(str(value)).quantize(Decimal(str(precision)))
        step = Decimal("1")
        if Decimal("10000") > abs(value) > Decimal("100"):
            step = Decimal("0.1")
        elif Decimal("100") > abs(value) > Decimal("1"):
            step = Decimal("0.01")
        elif Decimal("1") > abs(value) > Decimal("0.01"):
            step = Decimal("0.0001")
        elif Decimal("0.01") > abs(value) > Decimal("0.0001"):
            step = Decimal("0.00001")
        elif Decimal("0.0001") > abs(value) > s_decimal_0:
            step = Decimal("0.00000001")
        return (value // step) * step

    @staticmethod
    def divide(value, divisor):
        value = Decimal(str(value))
        divisor = Decimal(str(divisor))
        if divisor == s_decimal_0:
            return s_decimal_0
        return value / divisor

    def _is_trade_fill(self, trade):
        return type(trade) == TradeFill

    def _are_derivatives(self, trades: List[Any]) -> bool:
        return (
            trades
            and self._is_trade_fill(trades[0])
            and PositionAction.NIL.value not in [t.position for t in trades]
        )

    def _preprocess_trades_and_group_by_type(self, trades: List[Any]) -> Tuple[List[Any], List[Any]]:
        buys = []
        sells = []
        for trade in trades:
            if trade.trade_type.upper() == TradeType.BUY.name.upper():
                buys.append(trade)
                self.b_vol_base += Decimal(str(trade.amount))
                self.b_vol_quote += Decimal(str(trade.amount)) * Decimal(str(trade.price)) * Decimal("-1")
            elif trade.trade_type.upper() == TradeType.SELL.name.upper():
                sells.append(trade)
                self.s_vol_base += Decimal(str(trade.amount)) * Decimal("-1")
                self.s_vol_quote += Decimal(str(trade.amount)) * Decimal(str(trade.price))

            self.s_vol_quote += self._process_deducted_fees_impact_in_quote_vol(trade)

        self.tot_vol_base = self.b_vol_base + self.s_vol_base
        self.tot_vol_quote = self.b_vol_quote + self.s_vol_quote

        self.avg_b_price = self.divide(self.b_vol_quote, self.b_vol_base)
        self.avg_s_price = self.divide(self.s_vol_quote, self.s_vol_base)
        self.avg_tot_price = self.divide(abs(self.b_vol_quote) + abs(self.s_vol_quote),
                                         abs(self.b_vol_base) + abs(self.s_vol_base))
        self.avg_b_price = abs(self.avg_b_price)
        self.avg_s_price = abs(self.avg_s_price)

        return buys, sells

    def _process_deducted_fees_impact_in_quote_vol(self, trade):
        fee_percent = None
        fee_type = ""
        impact = s_decimal_0
        if self._is_trade_fill(trade):
            if trade.trade_fee.get("percent") is not None:
                fee_percent = Decimal(trade.trade_fee.get("percent"))
                fee_type = trade.trade_fee.get("fee_type")
        else:  # assume this is Trade object
            if trade.trade_fee.percent is not None:
                fee_percent = Decimal(trade.trade_fee.percent)
                fee_type = trade.trade_fee.type_descriptor_for_json()
        if (fee_percent is not None) and (fee_type == DeductedFromReturnsTradeFee.type_descriptor_for_json()):
            impact = Decimal(str(trade.amount)) * Decimal(str(trade.price)) * fee_percent * Decimal("-1")
        return impact

    async def _calculate_fees(self, quote: str, trades: List[Any]):
        for trade in trades:
            fee_percent = None
            trade_price = None
            trade_amount = None
            if self._is_trade_fill(trade):
                if trade.trade_fee.get("percent") is not None:
                    trade_price = Decimal(str(trade.price))
                    trade_amount = Decimal(str(trade.amount))
                    fee_percent = Decimal(str(trade.trade_fee["percent"]))
                flat_fees = [TokenAmount(token=flat_fee["token"], amount=Decimal(flat_fee["amount"]))
                             for flat_fee in trade.trade_fee.get("flat_fees", [])]
            else:  # assume this is Trade object
                if trade.trade_fee.percent is not None:
                    trade_price = Decimal(trade.price)
                    trade_amount = Decimal(trade.amount)
                    fee_percent = Decimal(trade.trade_fee.percent)
                flat_fees = trade.trade_fee.flat_fees

            if fee_percent is not None:
                self.fees[quote] += trade_price * trade_amount * fee_percent
            for flat_fee in flat_fees:
                self.fees[flat_fee.token] += flat_fee.amount

        for fee_token, fee_amount in self.fees.items():
            if fee_token == quote:
                self.fee_in_quote += fee_amount
            else:
                rate_pair: str = combine_to_hb_trading_pair(fee_token, quote)
                last_price = await RateOracle.get_instance().stored_or_live_rate(rate_pair)
                if last_price is not None:
                    self.fee_in_quote += fee_amount * last_price
                else:
                    self.logger().warning(
                        f"Could not find exchange rate for {rate_pair} "
                        f"using {RateOracle.get_instance()}. PNL value will be inconsistent."
                    )

    def _calculate_trade_pnl(self, buys: list, sells: list):
        self.trade_pnl = self.cur_value - self.hold_value

        # Handle trade_pnl differently for derivatives
        if self._are_derivatives(buys) or self._are_derivatives(sells):
            buys_copy, sells_copy = self.aggregate_position_order(buys.copy(), sells.copy())
            long = []
            short = []

            while True:
                lng = self.position_order(buys_copy, sells_copy)
                if lng is not None:
                    long.append(lng)

                sht = self.position_order(sells_copy, buys_copy)
                if sht is not None:
                    short.append(sht)
                if lng is None and sht is None:
                    break

            self.trade_pnl = Decimal(str(sum(self.derivative_pnl(long, short))))

    async def _initialize_metrics(self,
                                  trading_pair: str,
                                  trades: List[Any],
                                  current_balances: Dict[str, Decimal]):
        """
        Calculates PnL, fees, Return % and etc...
        :param trading_pair: the trading market to get performance metrics
        :param trades: the list of TradeFill or Trade object
        :param current_balances: current user account balance
        """

        base, quote = split_hb_trading_pair(trading_pair)
        buys, sells = self._preprocess_trades_and_group_by_type(trades)

        self.num_buys = len(buys)
        self.num_sells = len(sells)
        self.num_trades = self.num_buys + self.num_sells

        self.cur_base_bal = current_balances.get(base, s_decimal_0)
        self.cur_quote_bal = current_balances.get(quote, s_decimal_0)
        self.start_base_bal = self.cur_base_bal - self.tot_vol_base
        self.start_quote_bal = self.cur_quote_bal - self.tot_vol_quote

        self.start_price = Decimal(str(trades[0].price))
        self.cur_price = await RateOracle.get_instance().stored_or_live_rate(trading_pair)
        if self.cur_price is None:
            self.cur_price = Decimal(str(trades[-1].price))
        self.start_base_ratio_pct = self.divide(self.start_base_bal * self.start_price,
                                                (self.start_base_bal * self.start_price) + self.start_quote_bal)
        self.cur_base_ratio_pct = self.divide(self.cur_base_bal * self.cur_price,
                                              (self.cur_base_bal * self.cur_price) + self.cur_quote_bal)

        self.hold_value = (self.start_base_bal * self.cur_price) + self.start_quote_bal
        self.cur_value = (self.cur_base_bal * self.cur_price) + self.cur_quote_bal
        self._calculate_trade_pnl(buys, sells)

        await self._calculate_fees(quote, trades)

        self.total_pnl = self.trade_pnl - self.fee_in_quote
        self.return_pct = self.divide(self.total_pnl, self.hold_value)



import hashlib
import hmac
import uuid
from typing import Dict
from urllib.parse import urlencode, urlparse

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class BitstampAuth(AuthBase):
    AUTH_VERSION = "v2"

    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        local_headers = {}
        if request.headers is not None:
            local_headers.update(request.headers)

        auth_headers = self._generate_headers_for_authentication(
            request_url=request.url,
            content_type=local_headers.get("Content-Type"),
            payload=request.data,
            method=request.method,
        )

        local_headers.update(auth_headers)
        request.headers = local_headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. Bitstamp does not use this
        functionality
        """
        return request  # pass-through

    def _generate_headers_for_authentication(self, method: RESTMethod, request_url: str, content_type: str, payload) -> Dict[str, str]:
        nonce = str(uuid.uuid4())
        timestamp_str = str(int(self.time_provider.time() * 1e3))

        headers = {
            'X-Auth': 'BITSTAMP ' + self.api_key,
            'X-Auth-Signature': self._generate_signature(self._generate_message(method, request_url, content_type, payload, nonce, timestamp_str)),
            'X-Auth-Nonce': nonce,
            'X-Auth-Timestamp': timestamp_str,
            'X-Auth-Version': self.AUTH_VERSION
        }

        return headers

    def _generate_message(self, method: RESTMethod, request_url: str, content_type: str, payload, nonce: str, timestamp_str: str) -> str:
        content_type = content_type or ""
        payload_str = urlencode(payload) if payload else ""
        url = urlparse(request_url)
        message = f"BITSTAMP {self.api_key}{method}{url.hostname}{url.path}{content_type}{nonce}{timestamp_str}{self.AUTH_VERSION}{payload_str}"

        return message

    def _generate_signature(self, msg: str) -> str:
        digest = hmac.new(self.secret_key.encode("utf8"), msg=msg.encode("utf8"), digestmod=hashlib.sha256).hexdigest()
        return digest



import sys

import pyinjective.constant

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "injective_v2"

DEFAULT_DOMAIN = ""
TESTNET_DOMAIN = "testnet"

MAX_ORDER_ID_LEN = 36  # Injective supports uuid style client ids (36 characters)
HBOT_ORDER_ID_PREFIX = "HBOT"

DEFAULT_SUBACCOUNT_INDEX = 0
TX_GAS_PRICE = pyinjective.constant.GAS_PRICE
GAS_LIMIT_ADJUSTMENT_MULTIPLIER = None  # Leave as None to use the default value from the SDK. Otherwise, a float value.

EXPECTED_BLOCK_TIME = 1.5
TRANSACTIONS_CHECK_INTERVAL = 3 * EXPECTED_BLOCK_TIME
TRANSACTION_SUCCEEDED_CODE = 0

# Public limit ids
SPOT_MARKETS_LIMIT_ID = "SpotMarkets"
DERIVATIVE_MARKETS_LIMIT_ID = "DerivativeMarkets"
SPOT_ORDERBOOK_LIMIT_ID = "SpotOrderBookSnapshot"
DERIVATIVE_ORDERBOOK_LIMIT_ID = "DerivativeOrderBookSnapshot"
GET_TRANSACTION_LIMIT_ID = "GetTransaction"
FUNDING_RATES_LIMIT_ID = "FundingRates"
ORACLE_PRICES_LIMIT_ID = "OraclePrices"
FUNDING_PAYMENTS_LIMIT_ID = "FundingPayments"

# Private limit ids
PORTFOLIO_BALANCES_LIMIT_ID = "AccountPortfolio"
POSITIONS_LIMIT_ID = "Positions"
SPOT_ORDERS_HISTORY_LIMIT_ID = "SpotOrdersHistory"
DERIVATIVE_ORDERS_HISTORY_LIMIT_ID = "DerivativeOrdersHistory"
SPOT_TRADES_LIMIT_ID = "SpotTrades"
DERIVATIVE_TRADES_LIMIT_ID = "DerivativeTrades"
SIMULATE_TRANSACTION_LIMIT_ID = "SimulateTransaction"
SEND_TRANSACTION = "SendTransaction"

CHAIN_ENDPOINTS_GROUP_LIMIT_ID = "ChainGroupLimit"
INDEXER_ENDPOINTS_GROUP_LIMIT_ID = "IndexerGroupLimit"

NO_LIMIT = sys.maxsize
ONE_SECOND = 1

ENDPOINTS_RATE_LIMITS = [
    RateLimit(
        limit_id=SIMULATE_TRANSACTION_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(CHAIN_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=SEND_TRANSACTION,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(CHAIN_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=GET_TRANSACTION_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(CHAIN_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=SPOT_MARKETS_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=DERIVATIVE_MARKETS_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=SPOT_ORDERBOOK_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=DERIVATIVE_ORDERBOOK_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=PORTFOLIO_BALANCES_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=POSITIONS_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=SPOT_ORDERS_HISTORY_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=DERIVATIVE_ORDERS_HISTORY_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=SPOT_TRADES_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=DERIVATIVE_TRADES_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=FUNDING_RATES_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=ORACLE_PRICES_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
    RateLimit(
        limit_id=FUNDING_PAYMENTS_LIMIT_ID,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(INDEXER_ENDPOINTS_GROUP_LIMIT_ID)]),
]

PUBLIC_NODE_RATE_LIMITS = [
    RateLimit(limit_id=CHAIN_ENDPOINTS_GROUP_LIMIT_ID, limit=20, time_interval=ONE_SECOND),
    RateLimit(limit_id=INDEXER_ENDPOINTS_GROUP_LIMIT_ID, limit=50, time_interval=ONE_SECOND),
]
PUBLIC_NODE_RATE_LIMITS.extend(ENDPOINTS_RATE_LIMITS)

CUSTOM_NODE_RATE_LIMITS = [
    RateLimit(limit_id=CHAIN_ENDPOINTS_GROUP_LIMIT_ID, limit=NO_LIMIT, time_interval=ONE_SECOND),
    RateLimit(limit_id=INDEXER_ENDPOINTS_GROUP_LIMIT_ID, limit=NO_LIMIT, time_interval=ONE_SECOND),
]
CUSTOM_NODE_RATE_LIMITS.extend(ENDPOINTS_RATE_LIMITS)

ORDER_STATE_MAP = {
    "booked": OrderState.OPEN,
    "partial_filled": OrderState.PARTIALLY_FILLED,
    "filled": OrderState.FILLED,
    "canceled": OrderState.CANCELED,
}

STREAM_ORDER_STATE_MAP = {
    "Booked": OrderState.OPEN,
    "Matched": OrderState.FILLED,
    "Cancelled": OrderState.CANCELED,
}

ORDER_NOT_FOUND_ERROR_MESSAGE = "order not found"
ACCOUNT_SEQUENCE_MISMATCH_ERROR = "account sequence mismatch"

NEW_SPOT_ORDERS_EVENT_NAME = "injective.exchange.v1beta1.EventNewSpotOrders"
NEW_DERIVATIVE_ORDERS_EVENT_NAME = "injective.exchange.v1beta1.EventNewDerivativeOrders"



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.gateio.ws/api/v4"
HEALTH_CHECK_ENDPOINT = "/futures/usdt/contracts/BTC_USDT"
CANDLES_ENDPOINT = "/futures/usdt/candlesticks"
CONTRACT_INFO_URL = "/futures/usdt/contracts/{contract}"


WS_CANDLES_ENDPOINT = "futures.candlesticks"

WSS_URL = "wss://fx-ws.gateio.ws/v4/ws/usdt"
INTERVALS = bidict({
    "1m": "1m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1h",
    "4h": "4h",
    "8h": "8h",
    "1d": "1d",
    "7d": "7d",
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 2000
PUBLIC_URL_POINTS_LIMIT_ID = "PublicPoints"

RATE_LIMITS = [
    RateLimit(limit_id=PUBLIC_URL_POINTS_LIMIT_ID, limit=300, time_interval=1),
    RateLimit(limit_id=HEALTH_CHECK_ENDPOINT, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=CANDLES_ENDPOINT, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=CONTRACT_INFO_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
]



from typing import Dict, Type

from hummingbot.data_feed.candles_feed.ascend_ex_spot_candles.ascend_ex_spot_candles import AscendExSpotCandles
from hummingbot.data_feed.candles_feed.binance_perpetual_candles import BinancePerpetualCandles
from hummingbot.data_feed.candles_feed.binance_spot_candles import BinanceSpotCandles
from hummingbot.data_feed.candles_feed.bybit_perpetual_candles.bybit_perpetual_candles import BybitPerpetualCandles
from hummingbot.data_feed.candles_feed.bybit_spot_candles.bybit_spot_candles import BybitSpotCandles
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.data_feed.candles_feed.gate_io_perpetual_candles import GateioPerpetualCandles
from hummingbot.data_feed.candles_feed.gate_io_spot_candles import GateioSpotCandles
from hummingbot.data_feed.candles_feed.hyperliquid_perpetual_candles.hyperliquid_perpetual_candles import (
    HyperliquidPerpetualCandles,
)
from hummingbot.data_feed.candles_feed.hyperliquid_spot_candles.hyperliquid_spot_candles import HyperliquidSpotCandles
from hummingbot.data_feed.candles_feed.kraken_spot_candles.kraken_spot_candles import KrakenSpotCandles
from hummingbot.data_feed.candles_feed.kucoin_perpetual_candles.kucoin_perpetual_candles import KucoinPerpetualCandles
from hummingbot.data_feed.candles_feed.kucoin_spot_candles.kucoin_spot_candles import KucoinSpotCandles
from hummingbot.data_feed.candles_feed.mexc_perpetual_candles.mexc_perpetual_candles import MexcPerpetualCandles
from hummingbot.data_feed.candles_feed.mexc_spot_candles.mexc_spot_candles import MexcSpotCandles
from hummingbot.data_feed.candles_feed.okx_perpetual_candles.okx_perpetual_candles import OKXPerpetualCandles
from hummingbot.data_feed.candles_feed.okx_spot_candles.okx_spot_candles import OKXSpotCandles


class UnsupportedConnectorException(Exception):
    """
    Exception raised when an unsupported connector is requested.
    """
    def __init__(self, connector: str):
        message = f"The connector {connector} is not available. Please select another one."
        super().__init__(message)


class CandlesFactory:
    """
    The CandlesFactory class creates and returns a Candle object based on the specified configuration.
    It uses a mapping of connector names to their respective candle classes.
    """
    _candles_map: Dict[str, Type[CandlesBase]] = {
        "binance_perpetual": BinancePerpetualCandles,
        "binance": BinanceSpotCandles,
        "gate_io": GateioSpotCandles,
        "gate_io_perpetual": GateioPerpetualCandles,
        "kucoin": KucoinSpotCandles,
        "kucoin_perpetual": KucoinPerpetualCandles,
        "ascend_ex": AscendExSpotCandles,
        "okx_perpetual": OKXPerpetualCandles,
        "okx": OKXSpotCandles,
        "kraken": KrakenSpotCandles,
        "mexc": MexcSpotCandles,
        "mexc_perpetual": MexcPerpetualCandles,
        "bybit": BybitSpotCandles,
        "bybit_perpetual": BybitPerpetualCandles,
        "hyperliquid": HyperliquidSpotCandles,
        "hyperliquid_perpetual": HyperliquidPerpetualCandles
    }

    @classmethod
    def get_candle(cls, candles_config: CandlesConfig) -> CandlesBase:
        """
        Returns a Candle object based on the specified configuration.

        :param candles_config: CandlesConfig
        :return: Instance of CandleBase or its subclass.
        :raises UnsupportedConnectorException: If the connector is not supported.
        """
        connector_class = cls._candles_map.get(candles_config.connector)
        if connector_class:
            return connector_class(
                candles_config.trading_pair,
                candles_config.interval,
                candles_config.max_records
            )
        else:
            raise UnsupportedConnectorException(candles_config.connector)



from typing import Any, Callable, Dict, List, Optional

from hummingbot.connector.derivative.kucoin_perpetual import kucoin_perpetual_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.utils.tracking_nonce import get_tracking_nonce
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class HeadersContentRESTPreProcessor(RESTPreProcessorBase):
    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        request.headers = request.headers or {}
        request.headers["Content-Type"] = "application/json"
        return request


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        time_synchronizer: Optional[TimeSynchronizer] = None,
        time_provider: Optional[Callable] = None,
        auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time(throttler=throttler))
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(synchronizer=time_synchronizer, time_provider=time_provider),
            HeadersContentRESTPreProcessor(),
        ],
    )
    return api_factory


def create_throttler(trading_pairs: List[str] = None) -> AsyncThrottler:
    throttler = AsyncThrottler(CONSTANTS.RATE_LIMITS)
    return throttler


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None, domain: str = CONSTANTS.DEFAULT_DOMAIN
) -> float:
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    endpoint = CONSTANTS.SERVER_TIME_PATH_URL
    url = get_rest_url_for_endpoint(endpoint=endpoint, domain=domain)
    limit_id = get_rest_api_limit_id_for_endpoint(endpoint)
    response = await rest_assistant.execute_request(
        url=url,
        throttler_limit_id=limit_id,
        method=RESTMethod.GET,
    )
    server_time = response["data"]

    return server_time * 1e-3


def endpoint_from_message(message: Dict[str, Any]) -> Optional[str]:
    endpoint = None
    if "request" in message:
        message = message["request"]
    elif "type" in message:
        endpoint = message["type"]
    if isinstance(message, dict):
        if "subject" in message.keys():
            endpoint = message["subject"]
        elif endpoint is None and "topic" in message.keys():
            endpoint = message["topic"]
    return endpoint


def payload_from_message(message: Dict[str, Any]) -> Dict[str, Any]:
    payload = message
    if "data" in message:
        payload = message["data"]
    return payload


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def get_rest_url_for_endpoint(
    endpoint: str,
    domain: str = CONSTANTS.DEFAULT_DOMAIN
):
    variant = domain if domain else CONSTANTS.DEFAULT_DOMAIN
    return CONSTANTS.REST_URLS.get(variant) + endpoint


def get_pair_specific_limit_id(base_limit_id: str, trading_pair: str) -> str:
    limit_id = f"{base_limit_id}-{trading_pair}"
    return limit_id


def get_rest_api_limit_id_for_endpoint(endpoint: Dict[str, str]) -> str:
    return endpoint


def _wss_url(endpoint: Dict[str, str], connector_variant_label: Optional[str]) -> str:
    variant = connector_variant_label if connector_variant_label else CONSTANTS.DEFAULT_DOMAIN
    return endpoint.get(variant)


def wss_public_url(connector_variant_label: Optional[str]) -> str:
    return _wss_url(CONSTANTS.WSS_PUBLIC_URLS, connector_variant_label)


def wss_private_url(connector_variant_label: Optional[str]) -> str:
    return _wss_url(CONSTANTS.WSS_PRIVATE_URLS, connector_variant_label)


def next_message_id() -> str:
    return str(get_tracking_nonce())


async def api_request(path: str,
                      api_factory: Optional[WebAssistantsFactory] = None,
                      throttler: Optional[AsyncThrottler] = None,
                      domain: str = CONSTANTS.DEFAULT_DOMAIN,
                      params: Optional[Dict[str, Any]] = None,
                      data: Optional[Dict[str, Any]] = None,
                      method: RESTMethod = RESTMethod.GET,
                      is_auth_required: bool = False,
                      return_err: bool = False,
                      api_version: str = "v1",
                      limit_id: Optional[str] = None,
                      timeout: Optional[float] = None):

    throttler = throttler or create_throttler()

    api_factory = api_factory or build_api_factory()
    rest_assistant = await api_factory.get_rest_assistant()

    async with throttler.execute_task(limit_id=limit_id if limit_id else path):
        url = get_rest_url_for_endpoint(endpoint=path, domain=domain)

        request = RESTRequest(
            method=method,
            url=url,
            params=params,
            data=data,
            is_auth_required=is_auth_required,
            throttler_limit_id=limit_id if limit_id else path
        )
        response = await rest_assistant.call(request=request, timeout=timeout)

        if response.status != 200:
            if return_err:
                error_response = await response.json()
                return error_response
            else:
                error_response = await response.text()
                raise IOError(f"Error executing request {method.name} {path}. "
                              f"HTTP status is {response.status}. "
                              f"Error: {error_response}")
        return await response.json()



from decimal import Decimal
from typing import TYPE_CHECKING, Dict, Optional

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_gather

if TYPE_CHECKING:
    from hummingbot.connector.exchange.binance.binance_exchange import BinanceExchange


class BinanceRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._binance_exchange: Optional[BinanceExchange] = None  # delayed because of circular reference

    @property
    def name(self) -> str:
        return "binance"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        self._ensure_exchanges()
        results = {}
        tasks = [
            self._get_binance_prices(exchange=self._binance_exchange),
        ]
        task_results = await safe_gather(*tasks, return_exceptions=True)
        for task_result in task_results:
            if isinstance(task_result, Exception):
                self.logger().error(
                    msg="Unexpected error while retrieving rates from Binance. Check the log file for more info.",
                    exc_info=task_result,
                )
                break
            else:
                results.update(task_result)
        return results

    def _ensure_exchanges(self):
        if self._binance_exchange is None:
            self._binance_exchange = self._build_binance_connector_without_private_keys(domain="com")

    @staticmethod
    async def _get_binance_prices(exchange: 'BinanceExchange', quote_token: str = None) -> Dict[str, Decimal]:
        """
        Fetches binance prices

        :param exchange: The exchange instance from which to query prices.
        :param quote_token: A quote symbol, if specified only pairs with the quote symbol are included for prices
        :return: A dictionary of trading pairs and prices
        """
        pairs_prices = await exchange.get_all_pairs_prices()
        results = {}
        for pair_price in pairs_prices:
            try:
                trading_pair = await exchange.trading_pair_associated_to_exchange_symbol(symbol=pair_price["symbol"])
            except KeyError:
                continue  # skip pairs that we don't track
            if quote_token is not None:
                base, quote = split_hb_trading_pair(trading_pair=trading_pair)
                if quote != quote_token:
                    continue
            bid_price = pair_price.get("bidPrice")
            ask_price = pair_price.get("askPrice")
            if bid_price is not None and ask_price is not None and 0 < Decimal(bid_price) <= Decimal(ask_price):
                results[trading_pair] = (Decimal(bid_price) + Decimal(ask_price)) / Decimal("2")

        return results

    @staticmethod
    def _build_binance_connector_without_private_keys(domain: str) -> 'BinanceExchange':
        from hummingbot.client.hummingbot_application import HummingbotApplication
        from hummingbot.connector.exchange.binance.binance_exchange import BinanceExchange

        app = HummingbotApplication.main_application()
        client_config_map = app.client_config_map

        return BinanceExchange(
            client_config_map=client_config_map,
            binance_api_key="",
            binance_api_secret="",
            trading_pairs=[],
            trading_required=False,
            domain=domain,
        )



import asyncio
import logging
import time
from abc import ABCMeta
from typing import Any, Dict, Optional

from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger


class UserStreamTrackerDataSource(metaclass=ABCMeta):

    _logger: Optional[HummingbotLogger] = None

    def __init__(self):
        self._ws_assistant: Optional[WSAssistant] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    def last_recv_time(self) -> float:
        """
        Returns the time of the last received message

        :return: the timestamp of the last received message in seconds
        """
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def listen_for_user_stream(self, output: asyncio.Queue):
        """
        Connects to the user private channel in the exchange using a websocket connection. With the established
        connection listens to all balance events and order updates provided by the exchange, and stores them in the
        output queue

        :param output: the queue to use to store the received messages
        """
        while True:
            try:
                self._ws_assistant = await self._connected_websocket_assistant()
                await self._subscribe_channels(websocket_assistant=self._ws_assistant)
                await self._send_ping(websocket_assistant=self._ws_assistant)  # to update last_recv_timestamp
                await self._process_websocket_messages(websocket_assistant=self._ws_assistant, queue=output)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
            except Exception:
                self.logger().exception("Unexpected error while listening to user stream. Retrying after 5 seconds...")
                await self._sleep(1.0)
            finally:
                await self._on_user_stream_interruption(websocket_assistant=self._ws_assistant)
                self._ws_assistant = None

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange

        :return: an instance of WSAssistant connected to the exchange
        """
        raise NotImplementedError

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        raise NotImplementedError

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        async for ws_response in websocket_assistant.iter_messages():
            data = ws_response.data
            await self._process_event_message(event_message=data, queue=queue)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if len(event_message) > 0:
            queue.put_nowait(event_message)

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        websocket_assistant and await websocket_assistant.disconnect()

    async def _send_ping(self, websocket_assistant: WSAssistant):
        await websocket_assistant.ping()

    async def _sleep(self, delay: float):
        """
        Function added only to facilitate patching the sleep in unit tests without affecting the asyncio module

        :param delay: number of seconds to sleep
        """
        await asyncio.sleep(delay)

    def _time(self) -> float:
        return time.time()



from abc import ABC, abstractmethod

from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class AuthBase(ABC):
    """A base class for authentication objects that can be fed to the `WebAssistantsFactory`.

    Hint: If the authentication requires a simple REST request to acquire information from the
    server that is required in the message signature, this class can be passed a `RESTConnection`
    object that it can use to that end.
    """

    @abstractmethod
    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        ...

    @abstractmethod
    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        ...



import gzip
import io
import json
from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
EXAMPLE_PAIR = "AURA-USDT"
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.001"),
    taker_percent_fee_decimal=Decimal("0.001"),
    buy_percent_fee_deducted_from_returns=True
)


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("status") == 1


def decompress_ws_message(message):
    if isinstance(message, bytes):
        compressed_data = gzip.GzipFile(fileobj=io.BytesIO(message), mode='rb')
        decompressed_data = compressed_data.read()
        utf8_data = decompressed_data.decode('utf-8')
        return json.loads(utf8_data)
    else:
        return message


class BingXConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="bing_x", client_data=None)
    bingx_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your BingX API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    bingx_api_secret: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your BingX secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="bing_x")


KEYS = BingXConfigMap.construct()



import asyncio
from typing import TYPE_CHECKING, Any, Dict, Optional

from hummingbot.connector.exchange.okx import okx_constants as CONSTANTS
from hummingbot.connector.exchange.okx.okx_auth import OkxAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSPlainTextRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.okx.okx_exchange import OkxExchange


class OkxAPIUserStreamDataSource(UserStreamTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            auth: OkxAuth,
            connector: 'OkxExchange',
            api_factory: WebAssistantsFactory):
        super().__init__()
        self._auth: OkxAuth = auth
        self._connector = connector
        self._api_factory = api_factory

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """

        ws: WSAssistant = await self._get_ws_assistant()
        async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_CONNECTION_LIMIT_ID):
            await ws.connect(
                ws_url=CONSTANTS.OKX_WS_URI_PRIVATE,
                message_timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE)

        payload = {
            "op": "login",
            "args": [self._auth.websocket_login_parameters()]
        }

        login_request: WSJSONRequest = WSJSONRequest(payload=payload)

        async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_LOGIN_LIMIT_ID):
            await ws.send(login_request)

        response: WSResponse = await ws.receive()
        message = response.data
        if message.get("event") != "login":
            self.logger().error("Error authenticating the private websocket connection")
            raise IOError("Private websocket connection authentication failed")

        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            payload = {
                "op": "subscribe",
                "args": [{"channel": "account"}],
            }
            subscribe_account_request: WSJSONRequest = WSJSONRequest(payload=payload)

            payload = {
                "op": "subscribe",
                "args": [
                    {
                        "channel": "orders",
                        "instType": "SPOT",
                    }
                ]
            }
            subscribe_orders_request: WSJSONRequest = WSJSONRequest(payload=payload)

            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIPTION_LIMIT_ID):
                await websocket_assistant.send(subscribe_account_request)
            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIPTION_LIMIT_ID):
                await websocket_assistant.send(subscribe_orders_request)
            self.logger().info("Subscribed to private account and orders channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await super()._process_websocket_messages(
                    websocket_assistant=websocket_assistant,
                    queue=queue)
            except asyncio.TimeoutError:
                ping_request = WSPlainTextRequest(payload="ping")
                await websocket_assistant.send(request=ping_request)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if len(event_message) > 0 and "data" in event_message:
            queue.put_nowait(event_message)

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant



from .base_trailing_indicator import BaseTrailingIndicator
import numpy as np


class InstantVolatilityIndicator(BaseTrailingIndicator):
    def __init__(self, sampling_length: int = 30, processing_length: int = 15):
        super().__init__(sampling_length, processing_length)

    def _indicator_calculation(self) -> float:
        # The standard deviation should be calculated between ticks and not with a mean of the whole buffer
        # Otherwise if the asset is trending, changing the length of the buffer would result in a greater volatility as more ticks would be further away from the mean
        # which is a nonsense result. If volatility of the underlying doesn't change in fact, changing the length of the buffer shouldn't change the result.
        np_sampling_buffer = self._sampling_buffer.get_as_numpy_array()
        vol = np.sqrt(np.sum(np.square(np.diff(np_sampling_buffer))) / np_sampling_buffer.size)
        return vol

    def _processing_calculation(self) -> float:
        # Only the last calculated volatlity, not an average of multiple past volatilities
        return self._processing_buffer.get_last_value()



import logging
from typing import Dict, Optional

from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants import (
    WS_ORDER_SUBSCRIPTION_CHANNELS,
)
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils import (
    get_timestamp_from_exchange_time,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.logger import HummingbotLogger


class CoinbaseAdvancedTradeOrderBook(OrderBook):
    """
    Coinbase Advanced Trade Order Book class
    """
    # Mapping of WS channels to their respective sequence numbers
    _sequence_nums: Dict[str, int] = {channel: 0 for channel in WS_ORDER_SUBSCRIPTION_CHANNELS.inv.keys()}

    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger | logging.Logger:
        if cls._logger is None:
            name: str = HummingbotLogger.logger_name_for_class(cls)
            cls._logger = logging.getLogger(name)
        return cls._logger

    @classmethod
    def snapshot_message_from_exchange(cls,
                                       msg: Dict[str, any],
                                       timestamp: float,
                                       metadata: Optional[Dict] = None) -> OrderBookMessage:
        """
        Creates a snapshot message with the order book snapshot message
        :param msg: the response from the exchange when requesting the order book snapshot
        :param timestamp: the snapshot timestamp
        :param metadata: a dictionary with extra information to add to the snapshot data
        :return: a snapshot message with the snapshot information received from the exchange
        """
        if metadata:
            msg.update(metadata)
        return OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": msg["trading_pair"],
            "update_id": int(get_timestamp_from_exchange_time(msg["pricebook"]["time"], "s")),
            "bids": [[d["price"], d["size"]] for d in msg["pricebook"]["bids"]],
            "asks": [[d["price"], d["size"]] for d in msg["pricebook"]["asks"]]
        }, timestamp=timestamp)

    @classmethod
    def diff_message_from_exchange(
            cls,
            msg: Dict[str, any],
            timestamp: Optional[float] = None,
            metadata: Optional[Dict] = None) -> OrderBookMessage:
        """
        Process messages from the order book or trade channel
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#level2-channel
        The snapshot is the first message received form the 'level2' channel. It has a sequence_num = 0
        :param msg: the response from the exchange when requesting the order book snapshot
        :param symbol_to_pair: Method to retrieve a Hummingbot trading pair from an exchange symbol
        :return: a snapshot message with the snapshot information received from the exchange
        {
            'channel': 'l2_data',
            'client_id': '',
            'timestamp': '2024-10-08T09:10:36.04370306Z',
            'sequence_num': 9,
            'events': [
                {
                    'type': 'update',
                    'product_id': 'BTC-USD',
                    'updates': [
                        {
                            'side': 'bid',
                            'event_time': '2024-10-08T09:10:34.970831Z',
                            'price_level': '62319.46',
                            'new_quantity': '0.00169'
                        },
                        {
                            'side': 'bid',
                            'event_time': '2024-10-08T09:10:34.970831Z',
                            'price_level': '62318.81',
                            'new_quantity': '0'
                        },
                        {
                            'side': 'offer'
                            'event_time': '2024-10-08T09:10:34.970831Z',
                            'price_level': '62336.65',
                            'new_quantity': '0.03021537'
                        }
                    ]
                }
            ]
        }
        """
        if metadata:
            msg.update(metadata)
        event = msg["events"][0]
        # trading_pair = event["product_id"]
        obm_content = {
            "trading_pair": msg["trading_pair"],
            "update_id": int(get_timestamp_from_exchange_time(msg["timestamp"], "s")),
            "bids": [],
            "asks": []
        }
        for update in event.get("updates", []):
            if update["side"] == "bid":
                obm_content["bids"].append([update["price_level"], update["new_quantity"]])
            else:
                obm_content["asks"].append([update["price_level"], update["new_quantity"]])

        return OrderBookMessage(
            OrderBookMessageType.DIFF,
            obm_content,
            timestamp=obm_content['update_id'])

    @classmethod
    def trade_message_from_exchange(cls, msg: Dict[str, any], metadata: Optional[Dict] = None):
        """
        Process messages from the market trades channel
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#market-trades-channel
        :param msg: the response from the exchange when requesting the order book snapshot
        :param symbol_to_pair: Method to retrieve a Hummingbot trading pair from an exchange symbol
        :return: a trade message with the trade information received from the exchange
        {
            'channel': 'market_trades',
            'client_id': '',
            'timestamp': '2024-10-08T09:10:52.425044603Z',
            'sequence_num': 328,
            'events': [
                {
                    'type': 'update',
                    'trades': [
                        {
                            'product_id': 'BTC-USD',
                            'trade_id': '699871154',
                            'price': '62322.43',
                            'size': '0.00005535',
                            'time': '2024-10-08T09:10:52.378779Z',
                            'side': 'BUY'
                        }
                    ]
                }
            ]
        }
        """
        if metadata:
            msg.update(metadata)
        event = msg["events"][0]
        ts: float = get_timestamp_from_exchange_time(msg["timestamp"], "s")
        update = event["trades"][0]
        # Return an OrderBookMessage for each trade processed
        return OrderBookMessage(
            OrderBookMessageType.TRADE,
            {
                "trading_pair": msg["trading_pair"],
                "trade_type": float(TradeType.SELL.value) if update["side"] == "SELL" else float(TradeType.BUY.value),
                "trade_id": int(update["trade_id"]),
                "update_id": int(ts),
                "price": update["price"],
                "amount": update["size"]
            },
            timestamp=ts)



import asyncio
import logging
import time
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

import pandas as pd

from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import ClientConfigAdapter, get_connector_class
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import PriceType, TradeType
from hummingbot.core.data_type.order_book_query_result import OrderBookQueryResult
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.data_feed.candles_feed.candles_factory import CandlesFactory
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy_v2.executors.data_types import ConnectorPair


class MarketDataProvider:
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self,
                 connectors: Dict[str, ConnectorBase],
                 rates_update_interval: int = 60):
        self.candles_feeds = {}  # Stores instances of candle feeds
        self.connectors = connectors  # Stores instances of connectors
        self._rates_update_task = None
        self._rates_update_interval = rates_update_interval
        self._rates = {}
        self._rate_sources = {}
        self._rates_required = {}
        self.conn_settings = AllConnectorSettings.get_connector_settings()

    def stop(self):
        for candle_feed in self.candles_feeds.values():
            candle_feed.stop()
        if self._rates_update_task:
            self._rates_update_task.cancel()
            self._rates_update_task = None
        self.candles_feeds.clear()

    @property
    def ready(self) -> bool:
        # TODO: unify the ready property for connectors and feeds
        all_connectors_running = all(connector.ready for connector in self.connectors.values())
        all_candles_feeds_running = all(feed.ready for feed in self.candles_feeds.values())
        return all_connectors_running and all_candles_feeds_running

    def time(self):
        return time.time()

    def initialize_rate_sources(self, connector_pairs: List[ConnectorPair]):
        """
        Initializes a rate source based on the given connector pair.
        :param connector_pair: ConnectorPair
        """
        for connector_pair in connector_pairs:
            if connector_pair.is_amm_connector():
                if "gateway" not in self._rates_required:
                    self._rates_required["gateway"] = []
                self._rates_required["gateway"].append(connector_pair)
                continue
            if connector_pair.connector_name not in self._rates_required:
                self._rates_required[connector_pair.connector_name] = []
            self._rates_required[connector_pair.connector_name].append(connector_pair)
            if connector_pair.connector_name not in self._rate_sources:
                self._rate_sources[connector_pair.connector_name] = self.get_non_trading_connector(
                    connector_pair.connector_name)
        if not self._rates_update_task:
            self._rates_update_task = safe_ensure_future(self.update_rates_task())

    async def update_rates_task(self):
        """
        Updates the rates for all rate sources.
        """
        while True:
            rate_oracle = RateOracle.get_instance()
            for connector, connector_pairs in self._rates_required.items():
                if connector == "gateway":
                    tasks = []
                    gateway_client = GatewayHttpClient.get_instance()
                    for connector_pair in connector_pairs:
                        connector, chain, network = connector_pair.connector_name.split("_")
                        base, quote = connector_pair.trading_pair.split("-")
                        tasks.append(
                            gateway_client.get_price(
                                chain=chain, network=network, connector=connector,
                                base_asset=base, quote_asset=quote, amount=Decimal("1"),
                                side=TradeType.BUY))
                    try:
                        results = await asyncio.gather(*tasks)
                        for connector_pair, rate in zip(connector_pairs, results):
                            rate_oracle.set_price(connector_pair.trading_pair, Decimal(rate["price"]))
                    except Exception as e:
                        self.logger().error(f"Error fetching prices from {connector_pairs}: {e}", exc_info=True)
                else:
                    connector = self._rate_sources[connector]
                    prices = await self._safe_get_last_traded_prices(connector,
                                                                     [pair.trading_pair for pair in connector_pairs])
                    for pair, rate in prices.items():
                        rate_oracle.set_price(pair, rate)
            await asyncio.sleep(self._rates_update_interval)

    def initialize_candles_feed(self, config: CandlesConfig):
        """
        Initializes a candle feed based on the given configuration.
        :param config: CandlesConfig
        """
        self.get_candles_feed(config)

    def initialize_candles_feed_list(self, config_list: List[CandlesConfig]):
        """
        Initializes a list of candle feeds based on the given configurations.
        :param config_list: List[CandlesConfig]
        """
        for config in config_list:
            self.get_candles_feed(config)

    def get_candles_feed(self, config: CandlesConfig):
        """
        Retrieves or creates and starts a candle feed based on the given configuration.
        If an existing feed has a higher or equal max_records, it is reused.
        :param config: CandlesConfig
        :return: Candle feed instance.
        """
        key = self._generate_candle_feed_key(config)
        existing_feed = self.candles_feeds.get(key)

        if existing_feed and existing_feed.max_records >= config.max_records:
            # Existing feed is sufficient, return it
            return existing_feed
        else:
            # Create a new feed or restart the existing one with updated max_records
            candle_feed = CandlesFactory.get_candle(config)
            self.candles_feeds[key] = candle_feed
            if hasattr(candle_feed, 'start'):
                candle_feed.start()
            return candle_feed

    @staticmethod
    def _generate_candle_feed_key(config: CandlesConfig) -> str:
        """
        Generates a unique key for a candle feed based on its configuration.
        :param config: CandlesConfig
        :return: Unique key as a string.
        """
        return f"{config.connector}_{config.trading_pair}_{config.interval}"

    def stop_candle_feed(self, config: CandlesConfig):
        """
        Stops a candle feed based on the given configuration.
        :param config: CandlesConfig
        """
        key = self._generate_candle_feed_key(config)
        candle_feed = self.candles_feeds.get(key)
        if candle_feed and hasattr(candle_feed, 'stop'):
            candle_feed.stop()
            del self.candles_feeds[key]

    def get_connector(self, connector_name: str) -> ConnectorBase:
        """
        Retrieves a connector instance based on the given name.
        :param connector_name: str
        :return: ConnectorBase
        """
        connector = self.connectors.get(connector_name)
        if not connector:
            raise ValueError(f"Connector {connector_name} not found.")
        return connector

    def get_non_trading_connector(self, connector_name: str):
        conn_setting = self.conn_settings.get(connector_name)
        if conn_setting is None:
            self.logger().error(f"Connector {connector_name} not found")
            raise ValueError(f"Connector {connector_name} not found")

        client_config_map = ClientConfigAdapter(ClientConfigMap())
        connector_config = AllConnectorSettings.get_connector_config_keys(connector_name)
        api_keys = {key: "" for key in connector_config.__fields__.keys() if key != "connector"}
        init_params = conn_setting.conn_init_parameters(
            trading_pairs=[],
            trading_required=False,
            api_keys=api_keys,
            client_config_map=client_config_map,
        )
        connector_class = get_connector_class(connector_name)
        connector = connector_class(**init_params)
        return connector

    def get_balance(self, connector_name: str, asset: str):
        connector = self.get_connector(connector_name)
        return connector.get_balance(asset)

    def get_order_book(self, connector_name: str, trading_pair: str):
        """
        Retrieves the order book for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :return: Order book instance.
        """
        connector = self.get_connector(connector_name)
        return connector.get_order_book(trading_pair)

    def get_price_by_type(self, connector_name: str, trading_pair: str, price_type: PriceType):
        """
        Retrieves the price for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :param price_type: str
        :return: Price instance.
        """
        connector = self.get_connector(connector_name)
        return connector.get_price_by_type(trading_pair, price_type)

    def get_candles_df(self, connector_name: str, trading_pair: str, interval: str, max_records: int = 500):
        """
        Retrieves the candles for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :param interval: str
        :param max_records: int
        :return: Candles dataframe.
        """
        candles = self.get_candles_feed(CandlesConfig(
            connector=connector_name,
            trading_pair=trading_pair,
            interval=interval,
            max_records=max_records,
        ))
        return candles.candles_df.iloc[-max_records:]

    def get_trading_pairs(self, connector_name: str):
        """
        Retrieves the trading pairs from the specified connector.
        :param connector_name: str
        :return: List of trading pairs.
        """
        connector = self.get_connector(connector_name)
        return connector.trading_pairs

    def get_trading_rules(self, connector_name: str, trading_pair: str):
        """
        Retrieves the trading rules from the specified connector.
        :param connector_name: str
        :return: Trading rules.
        """
        connector = self.get_connector(connector_name)
        return connector.trading_rules[trading_pair]

    def quantize_order_price(self, connector_name: str, trading_pair: str, price: Decimal):
        connector = self.get_connector(connector_name)
        return connector.quantize_order_price(trading_pair, price)

    def quantize_order_amount(self, connector_name: str, trading_pair: str, amount: Decimal):
        connector = self.get_connector(connector_name)
        return connector.quantize_order_amount(trading_pair, amount)

    def get_price_for_volume(self, connector_name: str, trading_pair: str, volume: float,
                             is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the price for a specified volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param volume: The volume for which to find the price.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """

        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.get_price_for_volume(is_buy, volume)

    def get_order_book_snapshot(self, connector_name, trading_pair) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Retrieves the order book snapshot for a trading pair from the specified connector, as a tuple of bid and ask in
        DataFrame format.
        :param connector_name: str
        :param trading_pair: str
        :return: Tuple of bid and ask in DataFrame format.
        """
        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.snapshot

    def get_price_for_quote_volume(self, connector_name: str, trading_pair: str, quote_volume: float,
                                   is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the price for a specified quote volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param quote_volume: The quote volume for which to find the price.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.get_price_for_quote_volume(is_buy, quote_volume)

    def get_volume_for_price(self, connector_name: str, trading_pair: str, price: float,
                             is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the volume for a specified price on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param price: The price for which to find the volume.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.get_volume_for_price(is_buy, price)

    def get_quote_volume_for_price(self, connector_name: str, trading_pair: str, price: float,
                                   is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the quote volume for a specified price on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param price: The price for which to find the quote volume.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.get_quote_volume_for_price(is_buy, price)

    def get_vwap_for_volume(self, connector_name: str, trading_pair: str, volume: float,
                            is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the VWAP (Volume Weighted Average Price) for a specified volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param volume: The volume for which to calculate the VWAP.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        order_book = self.get_order_book(connector_name, trading_pair)
        return order_book.get_vwap_for_volume(is_buy, volume)

    def get_rate(self, pair: str) -> Decimal:
        """
        Finds a conversion rate for a given trading pair, this can be direct or indirect prices as
        long as it can find a route to achieve this.

        :param pair: A trading pair, e.g. BTC-USDT
        :return A conversion rate
        """
        return RateOracle.get_instance().get_pair_rate(pair)

    async def _safe_get_last_traded_prices(self, connector, trading_pairs, timeout=5):
        try:
            last_traded = await connector.get_last_traded_prices(trading_pairs=trading_pairs)
            return {pair: Decimal(rate) for pair, rate in last_traded.items()}
        except Exception as e:
            logging.error(
                f"Error getting last traded prices in connector {connector} for trading pairs {trading_pairs}: {e}")
            return {}



import asyncio
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from hummingbot.connector.derivative.kucoin_perpetual import (
    kucoin_perpetual_constants as CONSTANTS,
    kucoin_perpetual_web_utils as web_utils,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.derivative.kucoin_perpetual.kucoin_perpetual_derivative import KucoinPerpetualDerivative


class KucoinPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'KucoinPerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._nonce_provider = NonceCreator.for_microseconds()

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        funding_info_response = await self._request_complete_funding_info(trading_pair)
        if "symbol" in funding_info_response["data"]:
            symbol_info = funding_info_response["data"]
        else:
            symbol_info = funding_info_response["data"][0]
        funding_info = FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(str(symbol_info["indexPrice"])),
            mark_price=Decimal(str(symbol_info["markPrice"])),
            next_funding_utc_timestamp=int(pd.Timestamp(symbol_info["nextFundingRateTime"]).timestamp()),
            rate=Decimal(str(symbol_info["predictedFundingFeeRate"])),
        )
        return funding_info

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            symbols = ",".join([await self._connector.exchange_symbol_associated_to_pair(trading_pair=pair)
                                for pair in self._trading_pairs])

            trades_payload = {
                "id": web_utils.next_message_id(),
                "type": "subscribe",
                "topic": f"/contractMarket/ticker:{symbols}",
                "privateChannel": False,
                "response": False,
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            order_book_payload = {
                "id": web_utils.next_message_id(),
                "type": "subscribe",
                "topic": f"/contractMarket/level2:{symbols}",
                "privateChannel": False,
                "response": False,
            }
            subscribe_orderbook_request = WSJSONRequest(payload=order_book_payload)

            instrument_payload = {
                "id": web_utils.next_message_id(),
                "type": "subscribe",
                "topic": f"/contract/instrument:{symbols}",
                "privateChannel": False,
                "response": False,
            }
            subscribe_instruments_request = WSJSONRequest(payload=instrument_payload)
            await ws.send(subscribe_trade_request)  # not rate-limited
            await ws.send(subscribe_orderbook_request)  # not rate-limited
            await ws.send(subscribe_instruments_request)  # not rate-limited
            self.logger().info("Subscribed to public order book, trade and funding info channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        while True:
            try:
                await asyncio.wait_for(super()._process_websocket_messages(websocket_assistant=websocket_assistant),
                                       timeout=CONSTANTS.WS_CONNECTION_TIME_INTERVAL)
            except asyncio.TimeoutError:
                payload = {
                    "id": web_utils.next_message_id(),
                    "type": "ping",
                }
                ping_request = WSJSONRequest(payload=payload)
                self._last_ws_message_sent_timestamp = self._time()
                await websocket_assistant.send(request=ping_request)

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "data" in event_message and event_message.get("type") == "message":
            event_channel = event_message.get("topic")
            if CONSTANTS.WS_TRADES_TOPIC in event_channel:
                channel = self._trade_messages_queue_key
            elif CONSTANTS.WS_ORDER_BOOK_EVENTS_TOPIC in event_channel:
                channel = self._diff_messages_queue_key
            elif CONSTANTS.WS_INSTRUMENTS_INFO_TOPIC in event_channel:
                channel = self._funding_info_messages_queue_key
        return channel

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):

        event_type = raw_message["type"]

        if event_type == "message":
            symbol = raw_message["topic"].split(":")[-1]
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)
            diffs_data = raw_message["data"]
            timestamp: float = float(diffs_data["timestamp"]) * 1e-3
            bids = []
            asks = []
            price = diffs_data["change"].split(",")[0]
            side = diffs_data["change"].split(",")[1]
            quantity = Decimal(diffs_data["change"].split(",")[2])
            row_tuple = (price, self._connector.get_value_of_contracts(trading_pair, quantity))
            if side == "buy":
                bids.append(row_tuple)
            else:
                asks.append(row_tuple)
            order_book_message_content = {
                "trading_pair": trading_pair,
                "update_id": diffs_data["sequence"],
                "bids": bids,
                "asks": asks,
            }
            diff_message = OrderBookMessage(
                message_type=OrderBookMessageType.DIFF,
                content=order_book_message_content,
                timestamp=timestamp,
            )
            message_queue.put_nowait(diff_message)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trade_data: Dict[str, Any] = raw_message["data"]
        timestamp: float = int(trade_data["time"]) * 1e-9
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=trade_data["symbol"])
        message_content = {
            "trade_id": str(trade_data["tradeId"]),
            "update_id": int(trade_data["sequence"]),
            "trading_pair": trading_pair,
            "trade_type": float(TradeType.BUY.value) if trade_data["side"] == "buy" else float(
                TradeType.SELL.value),
            "amount": self._connector.get_value_of_contracts(trading_pair, Decimal(trade_data["size"])),
            "price": Decimal(trade_data["price"])
        }
        trade_message: Optional[OrderBookMessage] = OrderBookMessage(
            message_type=OrderBookMessageType.TRADE,
            content=message_content,
            timestamp=timestamp)

        message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        event_type = raw_message["subject"]
        if event_type == "funding.rate" or event_type == "mark.index.price" or event_type == "position.settlement":
            symbol = raw_message["topic"].split(":")[-1]
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)
            entries = raw_message["data"]
            info_update = FundingInfoUpdate(trading_pair)
            if "indexPrice" in entries:
                info_update.index_price = Decimal(str(entries["indexPrice"]))
            if "markPrice" in entries:
                info_update.mark_price = Decimal(str(entries["markPrice"]))
            if "fundingRate" in entries:
                info_update.rate = Decimal(str(entries["fundingRate"]))
            message_queue.put_nowait(info_update)

    async def _request_complete_funding_info(self, trading_pair: str) -> Dict[str, Any]:
        exchange_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
        rest_assistant = await self._api_factory.get_rest_assistant()
        endpoint = CONSTANTS.GET_CONTRACT_INFO_PATH_URL.format(symbol=exchange_symbol[0])
        url = web_utils.get_rest_url_for_endpoint(endpoint=endpoint, domain=self._domain)
        data = await rest_assistant.execute_request(
            url=url,
            throttler_limit_id=CONSTANTS.GET_CONTRACT_INFO_PATH_URL,
            method=RESTMethod.GET,
            is_auth_required=True,
        )
        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        exchange_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
        if len(exchange_symbol) > 0:
            exchange_symbol = exchange_symbol[0]
        snapshot_response = await self._request_order_book_snapshot(exchange_symbol)
        snapshot_data = snapshot_response["data"]
        if "time" in snapshot_data:
            timestamp = float(snapshot_data["time"]) * 1e-3
        elif "ts" in snapshot_data:
            timestamp = float(snapshot_data["ts"]) * 1e-9
        else:
            timestamp = time.time()
        if "sequence" in snapshot_data:
            update_id = int(snapshot_data["sequence"])
        else:
            update_id = self._nonce_provider.get_tracking_nonce(timestamp=timestamp)
        bids, asks = self._get_bids_and_asks_from_rest_msg_data(trading_pair, snapshot_data)
        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": bids,
            "asks": asks,
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            message_type=OrderBookMessageType.SNAPSHOT,
            content=order_book_message_content,
            timestamp=timestamp,
        )

        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        rest_assistant = await self._api_factory.get_rest_assistant()
        endpoint = CONSTANTS.ORDER_BOOK_ENDPOINT
        url = web_utils.get_rest_url_for_endpoint(endpoint=endpoint.format(symbol=trading_pair))
        limit_id = web_utils.get_rest_api_limit_id_for_endpoint(endpoint)
        data = await rest_assistant.execute_request(
            url=url,
            throttler_limit_id=limit_id,
            method=RESTMethod.GET,
        )

        return data

    def _get_bids_and_asks_from_rest_msg_data(
            self, trading_pair, snapshot: List[Dict[str, Union[str, int, float]]]
    ) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:
        bids = [
            (float(row[0]), self._connector.get_value_of_contracts(trading_pair, Decimal(row[1])))
            for row in snapshot['bids']
        ]
        asks = [
            (float(row[0]), self._connector.get_value_of_contracts(trading_pair, Decimal(row[1])))
            for row in snapshot['asks']
        ]
        return bids, asks

    @staticmethod
    def _get_bids_and_asks_from_ws_msg_data(
            snapshot: Dict[str, List[Dict[str, Union[str, int, float]]]]
    ) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:
        bids = []
        asks = []
        for action, rows_list in snapshot.items():
            if action not in ["delete", "update", "insert"]:
                continue
            is_delete = action == "delete"
            for row_dict in rows_list:
                row_price = row_dict["price"]
                row_size = 0.0 if is_delete else row_dict["size"]
                row_tuple = (row_price, row_size)
                if row_dict["side"] == "Buy":
                    bids.append(row_tuple)
                else:
                    asks.append(row_tuple)
        return bids, asks

    async def _connected_websocket_assistant(self) -> WSAssistant:
        rest_assistant = await self._api_factory.get_rest_assistant()
        connection_info = await rest_assistant.execute_request(
            url=web_utils.get_rest_url_for_endpoint(endpoint=CONSTANTS.PUBLIC_WS_DATA_PATH_URL, domain=self._domain),
            method=RESTMethod.POST,
            throttler_limit_id=CONSTANTS.PUBLIC_WS_DATA_PATH_URL,
        )

        ws_url = connection_info["data"]["instanceServers"][0]["endpoint"]
        self._ping_interval = int(connection_info["data"]["instanceServers"][0]["pingInterval"]) * 0.8 * 1e-3
        # message_timeout = int(connection_info["data"]["instanceServers"][0]["pingTimeout"]) * 0.8 * 1e-3
        token = connection_info["data"]["token"]

        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=f"{ws_url}?token={token}", ping_timeout=self._ping_interval)
        # await ws.connect(ws_url=f"{ws_url}?token={token}", ping_timeout=self._ping_interval, message_timeout=message_timeout)
        return ws



from base_trailing_indicator import BaseTrailingIndicator
import pandas as pd


class ExponentialMovingAverageIndicator(BaseTrailingIndicator):
    def __init__(self, sampling_length: int = 30, processing_length: int = 1):
        if processing_length != 1:
            raise Exception("Exponential moving average processing_length should be 1")
        super().__init__(sampling_length, processing_length)

    def _indicator_calculation(self) -> float:
        ema = pd.Series(self._sampling_buffer.get_as_numpy_array())\
            .ewm(span=self._sampling_length, adjust=True).mean()
        return ema[-1]

    def _processing_calculation(self) -> float:
        return self._processing_buffer.get_last_value()



from decimal import Decimal
from typing import Any, Dict

from pydantic import SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0008"),
    taker_percent_fee_decimal=Decimal("0.001"),
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USDT"


class OKXConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="okx", const=True, client_data=None)
    okx_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your OKX API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )
    okx_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your OKX secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )
    okx_passphrase: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your OKX passphrase key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )


KEYS = OKXConfigMap.construct()


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("instType", None) == "SPOT"



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.derivative.bitget_perpetual import bitget_perpetual_constants as CONSTANTS
from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_auth import BitgetPerpetualAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSPlainTextRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_derivative import BitgetPerpetualDerivative


class BitgetPerpetualUserStreamDataSource(UserStreamTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        auth: BitgetPerpetualAuth,
        trading_pairs: List[str],
        connector: 'BitgetPerpetualDerivative',
        api_factory: WebAssistantsFactory,
        domain: str = None,
    ):
        super().__init__()
        self._domain = domain
        self._api_factory = api_factory
        self._auth = auth
        self._trading_pairs = trading_pairs
        self._connector = connector
        self._pong_response_event = None

    async def _authenticate(self, ws: WSAssistant):
        """
        Authenticates user to websocket
        """
        auth_payload: List[str] = self._auth.get_ws_auth_payload()
        payload = {"op": "login", "args": auth_payload}
        login_request: WSJSONRequest = WSJSONRequest(payload=payload)
        await ws.send(login_request)
        response: WSResponse = await ws.receive()
        message = response.data

        if (
            message["event"] != "login"
            and message["code"] != "0"
        ):
            self.logger().error("Error authenticating the private websocket connection")
            raise IOError("Private websocket connection authentication failed")

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant, queue=queue),
                    timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE)
            except asyncio.TimeoutError:
                if self._pong_response_event and not self._pong_response_event.is_set():
                    # The PONG response for the previous PING request was never received
                    raise IOError("The user stream channel is unresponsive (pong response not received)")
                self._pong_response_event = asyncio.Event()
                await self._send_ping(websocket_assistant=websocket_assistant)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if event_message == CONSTANTS.WS_PONG_RESPONSE and self._pong_response_event:
            self._pong_response_event.set()
        elif "event" in event_message:
            if event_message["event"] == "error":
                raise IOError(f"Private channel subscription failed ({event_message})")
        else:
            await super()._process_event_message(event_message=event_message, queue=queue)

    async def _send_ping(self, websocket_assistant: WSAssistant):
        ping_request = WSPlainTextRequest(payload=CONSTANTS.WS_PING_REQUEST)
        await websocket_assistant.send(ping_request)

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            product_types = set([await self._connector.product_type_for_trading_pair(trading_pair=trading_pair)
                                 for trading_pair in self._trading_pairs])
            subscription_payloads = []

            for product_type in product_types:
                subscription_payloads.append(
                    {
                        "instType": product_type.upper(),
                        "channel": CONSTANTS.WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME,
                        "instId": "default"
                    }
                )
                subscription_payloads.append(
                    {
                        "instType": product_type.upper(),
                        "channel": CONSTANTS.WS_SUBSCRIPTION_POSITIONS_ENDPOINT_NAME,
                        "instId": "default"
                    }
                )
                subscription_payloads.append(
                    {
                        "instType": product_type.upper(),
                        "channel": CONSTANTS.WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME,
                        "instId": "default"
                    }
                )

            payload = {
                "op": "subscribe",
                "args": subscription_payloads
            }
            subscription_request = WSJSONRequest(payload)

            await websocket_assistant.send(subscription_request)

            self.logger().info("Subscribed to private account, position and orders channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(
                "Unexpected error occurred subscribing to account, position and orders channels..."
            )
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(
            ws_url=CONSTANTS.WSS_URL,
            message_timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE)
        await self._authenticate(ws)
        return ws



from copy import deepcopy
from typing import AsyncGenerator, Dict, List, Optional

from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import WSRequest, WSResponse
from hummingbot.core.web_assistant.connections.ws_connection import WSConnection
from hummingbot.core.web_assistant.ws_post_processors import WSPostProcessorBase
from hummingbot.core.web_assistant.ws_pre_processors import WSPreProcessorBase


class WSAssistant:
    """A helper class to contain all WebSocket-related logic.

    The class can be injected with additional functionality by passing a list of objects inheriting from
    the `WSPreProcessorBase` and `WSPostProcessorBase` classes. The pre-processors are applied to a request
    before it is sent out, while the post-processors are applied to a response before it is returned to the caller.
    """

    def __init__(
        self,
        connection: WSConnection,
        ws_pre_processors: Optional[List[WSPreProcessorBase]] = None,
        ws_post_processors: Optional[List[WSPostProcessorBase]] = None,
        auth: Optional[AuthBase] = None,
    ):
        self._connection = connection
        self._ws_pre_processors = ws_pre_processors or []
        self._ws_post_processors = ws_post_processors or []
        self._auth = auth

    @property
    def last_recv_time(self) -> float:
        return self._connection.last_recv_time

    async def connect(
        self,
        ws_url: str,
        *,
        ping_timeout: float = 10,
        message_timeout: Optional[float] = None,
        ws_headers: Optional[Dict] = {},
        max_msg_size: Optional[int] = None,
    ):
        max_msg_size = max_msg_size if max_msg_size else self._connection._MAX_MSG_SIZE
        await self._connection.connect(
            ws_url=ws_url,
            ws_headers=ws_headers,
            ping_timeout=ping_timeout,
            message_timeout=message_timeout,
            max_msg_size=max_msg_size)

    async def disconnect(self):
        await self._connection.disconnect()

    async def subscribe(self, request: WSRequest):
        """Will eventually be used to handle automatic re-connection."""
        await self.send(request)

    async def send(self, request: WSRequest):
        request = deepcopy(request)
        request = await self._pre_process_request(request)
        request = await self._authenticate(request)
        await self._connection.send(request)

    async def ping(self):
        await self._connection.ping()

    async def iter_messages(self) -> AsyncGenerator[Optional[WSResponse], None]:
        """Will yield None and stop if `WSDelegate.disconnect()` is called while waiting for a response."""
        while self._connection.connected:
            response = await self._connection.receive()
            if response is not None:
                response = await self._post_process_response(response)
                yield response

    async def receive(self) -> Optional[WSResponse]:
        """This method will return `None` if `WSDelegate.disconnect()` is called while waiting for a response."""
        response = await self._connection.receive()
        if response is not None:
            response = await self._post_process_response(response)
        return response

    async def _pre_process_request(self, request: WSRequest) -> WSRequest:
        for pre_processor in self._ws_pre_processors:
            request = await pre_processor.pre_process(request)
        return request

    async def _authenticate(self, request: WSRequest) -> WSRequest:
        if self._auth is not None and request.is_auth_required:
            request = await self._auth.ws_authenticate(request)
        return request

    async def _post_process_response(self, response: WSResponse) -> WSResponse:
        for post_processor in self._ws_post_processors:
            response = await post_processor.post_process(response)
        return response



import json
import time
from abc import ABC, abstractmethod
from typing import Optional

from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS
from hummingbot.connector.utilities.oms_connector.oms_connector_auth import OMSConnectorAuth
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.connections.data_types import WSRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_post_processors import WSPostProcessorBase
from hummingbot.core.web_assistant.ws_pre_processors import WSPreProcessorBase


class OMSConnectorURLCreatorBase(ABC):
    @abstractmethod
    def get_rest_url(self, path_url: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_ws_url(self) -> str:
        raise NotImplementedError


class OMSConnectorWSPreProcessor(WSPreProcessorBase):
    def __init__(self):
        self._msg_sequence_num = 0

    async def pre_process(self, request: WSRequest) -> WSRequest:
        request.payload[CONSTANTS.MSG_TYPE_FIELD] = CONSTANTS.REQ_MSG_TYPE
        sequence_num = self._generate_sequence_number()
        request.payload[CONSTANTS.MSG_SEQUENCE_FIELD] = sequence_num
        request.payload[CONSTANTS.MSG_DATA_FIELD] = json.dumps(request.payload[CONSTANTS.MSG_DATA_FIELD])
        return request

    def _generate_sequence_number(self) -> int:
        self._msg_sequence_num += 2
        return self._msg_sequence_num


class OMSConnectorWSPostProcessor(WSPostProcessorBase):
    async def post_process(self, response: WSResponse) -> WSResponse:
        if CONSTANTS.MSG_DATA_FIELD in response.data:
            response.data[CONSTANTS.MSG_DATA_FIELD] = json.loads(response.data[CONSTANTS.MSG_DATA_FIELD])
        return response


class OMSConnectorWebAssistantsFactory(WebAssistantsFactory):
    @property
    def auth(self) -> Optional[OMSConnectorAuth]:
        return self._auth


def build_api_factory(
    throttler: Optional[AsyncThrottler] = None,
    auth: Optional[OMSConnectorAuth] = None,
):
    throttler = throttler or create_throttler()
    api_factory = OMSConnectorWebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        ws_pre_processors=[OMSConnectorWSPreProcessor()],
        ws_post_processors=[OMSConnectorWSPostProcessor()],
    )
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None, domain: str = ""
) -> float:
    return _time() * 1e3


def _time() -> float:
    """Can be mocked in unit-tests without directly affecting `time.time()`"""
    return time.time()



import asyncio
import copy
import importlib
import inspect
import json
import os
import shutil
import sys
from collections import OrderedDict
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional

import yaml
from pydantic.v1.fields import FieldInfo

from hummingbot.client import settings
from hummingbot.client.config.config_data_types import BaseClientModel
from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    ConfigValidationError,
    default_strategy_file_path,
    format_config_file_name,
    get_strategy_config_map,
    get_strategy_template_path,
    parse_config_default_to_text,
    parse_cvar_value,
    save_previous_strategy_value,
    save_to_yml,
    save_to_yml_legacy,
)
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.config.strategy_config_data_types import BaseStrategyConfigMap
from hummingbot.client.settings import SCRIPT_STRATEGY_CONF_DIR_PATH, STRATEGIES_CONF_DIR_PATH, required_exchanges
from hummingbot.client.ui.completer import load_completer
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.exceptions import InvalidController, InvalidScriptModule
from hummingbot.strategy.strategy_v2_base import StrategyV2ConfigBase
from hummingbot.strategy_v2.controllers.controller_base import ControllerConfigBase
from hummingbot.strategy_v2.controllers.directional_trading_controller_base import (
    DirectionalTradingControllerConfigBase,
)
from hummingbot.strategy_v2.controllers.market_making_controller_base import MarketMakingControllerConfigBase

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class OrderedDumper(yaml.SafeDumper):
    pass


class CreateCommand:
    def create(self,  # type: HummingbotApplication
               script_to_config: Optional[str] = None,
               controller_name: Optional[str] = None, ) -> None:
        self.app.clear_input()
        self.placeholder_mode = True
        self.app.hide_input = True
        required_exchanges.clear()
        if script_to_config and controller_name:
            self.notify("Please provide only one of script or controller name.")
            return
        if script_to_config:
            safe_ensure_future(self.prompt_for_configuration_v2(script_to_config))
        elif controller_name:
            safe_ensure_future(self.prompt_for_controller_config(controller_name))
        else:
            safe_ensure_future(self.prompt_for_configuration())

    async def prompt_for_controller_config(self,  # type: HummingbotApplication
                                           controller_name: str):
        try:

            # Attempt to find and load the correct module
            module = None
            try:
                module_path = f"{settings.CONTROLLERS_MODULE}.{controller_name}"
                module = importlib.import_module(module_path)
            except ImportError:
                pass

            if not module:
                raise InvalidController(f"The controller {controller_name} was not found in any subfolder.")

            # Load the configuration class from the module
            config_class = next((member for member_name, member in inspect.getmembers(module)
                                 if inspect.isclass(member) and member not in [ControllerConfigBase,
                                                                               MarketMakingControllerConfigBase,
                                                                               DirectionalTradingControllerConfigBase,]
                                 and (issubclass(member, ControllerConfigBase))), None)
            if not config_class:
                raise InvalidController(f"No configuration class found in the module {controller_name}.")

            config_class_instance = config_class.construct()
            config_class_instance.id = config_class_instance.set_id(None)
            config_map = ClientConfigAdapter(config_class_instance)

            await self.prompt_for_model_config(config_map)
            if not self.app.to_stop_config:
                file_name = await self.save_config(controller_name, config_map, settings.CONTROLLERS_CONF_DIR_PATH)
                self.notify(f"A new config file has been created. Edit the file in your IDE to adjust the parameters: {file_name}")

            self.app.change_prompt(prompt=">>> ")
            self.app.input_field.completer = load_completer(self)
            self.placeholder_mode = False
            self.app.hide_input = False

        except StopIteration:
            raise InvalidController(f"The module {controller_name} does not contain any subclass of BaseModel")
        except Exception as e:
            self.notify(f"An error occurred: {str(e)}")
            self.reset_application_state()

    async def prompt_for_configuration_v2(self,  # type: HummingbotApplication
                                          script_to_config: str):
        try:
            module = sys.modules.get(f"{settings.SCRIPT_STRATEGIES_MODULE}.{script_to_config}")
            script_module = importlib.reload(module)
            config_class = next((member for member_name, member in inspect.getmembers(script_module)
                                 if
                                 inspect.isclass(member) and member not in [BaseClientModel, StrategyV2ConfigBase] and
                                 (issubclass(member, BaseClientModel) or issubclass(member, StrategyV2ConfigBase))))
            config_map = ClientConfigAdapter(config_class.model_construct())

            await self.prompt_for_model_config(config_map)
            if not self.app.to_stop_config:
                file_name = await self.save_config(script_to_config, config_map, SCRIPT_STRATEGY_CONF_DIR_PATH)
                self.notify(f"A new config file has been created: {file_name}")
            self.app.change_prompt(prompt=">>> ")
            self.app.input_field.completer = load_completer(self)
            self.placeholder_mode = False
            self.app.hide_input = False

        except StopIteration:
            raise InvalidScriptModule(f"The module {script_to_config} does not contain any subclass of BaseModel")
        except Exception as e:
            self.notify(f"An error occurred: {str(e)}")
            self.reset_application_state()

    async def save_config(self, name: str, config_instance: BaseClientModel, config_dir_path: Path):
        file_name = await self.prompt_new_file_name(name, True)
        if self.app.to_stop_config:
            self.app.set_text("")
            return
        config_path = config_dir_path / file_name

        # Check if the file already exists
        if config_path.exists():
            self.notify(f"File {file_name} already exists. Please enter a different file name.")
            return await self.save_config(name, config_instance, config_dir_path)  # Recursive call

        config_path = config_dir_path / file_name
        field_order = list(config_instance.model_fields.keys())
        config_json_str = config_instance.model_dump_json()
        config_data = json.loads(config_json_str)
        ordered_config_data = OrderedDict((field, config_data.get(field)) for field in field_order)

        def _dict_representer(dumper, data):
            return dumper.represent_dict(data.items())

        OrderedDumper.add_representer(OrderedDict, _dict_representer)
        with open(config_path, 'w') as file:
            yaml.dump(ordered_config_data, file, Dumper=OrderedDumper, default_flow_style=False)

        return file_name

    async def prompt_for_configuration(
        self,  # type: HummingbotApplication
    ):
        strategy = await self.get_strategy_name()

        if self.app.to_stop_config:
            return

        config_map = get_strategy_config_map(strategy)
        self.notify(f"Please see https://docs.hummingbot.org/strategies/{strategy.replace('_', '-')}/ "
                    f"while setting up these below configuration.")

        if isinstance(config_map, ClientConfigAdapter):
            await self.prompt_for_model_config(config_map)
            if not self.app.to_stop_config:
                file_name = await self.save_config_to_file(config_map)
        elif config_map is not None:
            file_name = await self.prompt_for_configuration_legacy(strategy, config_map)
        else:
            self.app.to_stop_config = True

        if self.app.to_stop_config:
            return

        save_previous_strategy_value(file_name, self.client_config_map)
        self.strategy_file_name = file_name
        self.strategy_name = strategy
        self.strategy_config_map = config_map
        # Reload completer here otherwise the new file will not appear
        self.app.input_field.completer = load_completer(self)
        self.notify(f"A new config file has been created: {self.strategy_file_name}")
        self.placeholder_mode = False
        self.app.hide_input = False

        await self.verify_status()

    async def get_strategy_name(
        self,  # type: HummingbotApplication
    ) -> Optional[str]:
        strategy = None
        strategy_config = ClientConfigAdapter(BaseStrategyConfigMap.construct())
        await self.prompt_for_model_config(strategy_config)
        if not self.app.to_stop_config:
            strategy = strategy_config.strategy
        return strategy

    async def prompt_for_model_config(
        self,  # type: HummingbotApplication
        config_map: ClientConfigAdapter,
    ):
        for key in config_map.keys():
            client_data = config_map.get_client_data(key)
            if (
                client_data is not None
                and (client_data.prompt_on_new or config_map.is_required(key))
            ):
                await self.prompt_a_config(config_map, key)
                if self.app.to_stop_config:
                    break
            elif isinstance(getattr(config_map, key), FieldInfo):
                # If the config is a FieldInfo and the other checks passed we use the default value
                default_value = getattr(config_map, key).default
                setattr(config_map, key, default_value)

    async def prompt_for_configuration_legacy(
        self,  # type: HummingbotApplication
        strategy: str,
        config_map: Dict,
    ):
        config_map_backup = copy.deepcopy(config_map)
        # assign default values and reset those not required
        for config in config_map.values():
            if config.required:
                config.value = config.default
            else:
                config.value = None
        for config in config_map.values():
            if config.prompt_on_new and config.required:
                if not self.app.to_stop_config:
                    await self.prompt_a_config_legacy(config)
                else:
                    break
            else:
                config.value = config.default

        if self.app.to_stop_config:
            self.restore_config_legacy(config_map, config_map_backup)
            self.app.set_text("")
            return

        file_name = await self.prompt_new_file_name(strategy)
        if self.app.to_stop_config:
            self.restore_config_legacy(config_map, config_map_backup)
            self.app.set_text("")
            return
        self.app.change_prompt(prompt=">>> ")
        strategy_path = STRATEGIES_CONF_DIR_PATH / file_name
        template = get_strategy_template_path(strategy)
        shutil.copy(template, strategy_path)
        save_to_yml_legacy(str(strategy_path), config_map)
        return file_name

    async def prompt_a_config(
        self,  # type: HummingbotApplication
        model: ClientConfigAdapter,
        config: str,
        input_value=None,
        assign_default=True,
    ):
        config_path = config.split(".")
        while len(config_path) != 1:
            sub_model_attr = config_path.pop(0)
            model = getattr(model, sub_model_attr)
        config = config_path[0]
        if input_value is None:
            prompt = await model.get_client_prompt(config)
            if prompt is not None:
                if assign_default:
                    default = model.get_default_str_repr(attr_name=config)
                    self.app.set_text(default)
                prompt = f"{prompt} >>> "
                client_data = model.get_client_data(config)
                input_value = await self.app.prompt(prompt=prompt, is_password=client_data.is_secure)

        new_config_value = None
        if not self.app.to_stop_config and input_value is not None:
            try:
                setattr(model, config, input_value)
                new_config_value = getattr(model, config)
            except ConfigValidationError as e:
                self.notify(str(e))
                new_config_value = await self.prompt_a_config(model, config)

        if not self.app.to_stop_config and isinstance(new_config_value, ClientConfigAdapter):
            await self.prompt_for_model_config(new_config_value)

    async def prompt_a_config_legacy(
        self,  # type: HummingbotApplication
        config: ConfigVar,
        input_value=None,
        assign_default=True,
    ):
        if config.key == "inventory_price":
            await self.inventory_price_prompt_legacy(self.strategy_config_map, input_value)
            return
        if input_value is None:
            if assign_default:
                self.app.set_text(parse_config_default_to_text(config))
            prompt = await config.get_prompt()
            input_value = await self.app.prompt(prompt=prompt, is_password=config.is_secure)

        if self.app.to_stop_config:
            return
        value = parse_cvar_value(config, input_value)
        err_msg = await config.validate(input_value)
        if err_msg is not None:
            self.notify(err_msg)
            config.value = None
            await self.prompt_a_config_legacy(config)
        else:
            config.value = value

    async def save_config_to_file(
        self,  # type: HummingbotApplication
        config_map: ClientConfigAdapter,
    ) -> str:
        file_name = await self.prompt_new_file_name(config_map.strategy)
        if self.app.to_stop_config:
            self.app.set_text("")
            return
        self.app.change_prompt(prompt=">>> ")
        strategy_path = Path(STRATEGIES_CONF_DIR_PATH) / file_name
        save_to_yml(strategy_path, config_map)
        return file_name

    async def prompt_new_file_name(self,  # type: HummingbotApplication
                                   strategy: str,
                                   is_script: bool = False):
        file_name = default_strategy_file_path(strategy)
        self.app.set_text(file_name)
        input = await self.app.prompt(prompt="Enter a new file name for your configuration >>> ")
        input = format_config_file_name(input)
        conf_dir_path = STRATEGIES_CONF_DIR_PATH if not is_script else SCRIPT_STRATEGY_CONF_DIR_PATH
        file_path = os.path.join(conf_dir_path, input)
        if input is None or input == "":
            self.notify("Value is required.")
            return await self.prompt_new_file_name(strategy, is_script)
        elif os.path.exists(file_path):
            self.notify(f"{input} file already exists, please enter a new name.")
            return await self.prompt_new_file_name(strategy, is_script)
        else:
            return input

    async def verify_status(
        self  # type: HummingbotApplication
    ):
        try:
            timeout = float(self.client_config_map.commands_timeout.create_command_timeout)
            all_status_go = await asyncio.wait_for(self.status_check_all(), timeout)
        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the connection check to complete. See logs for more details.")
            self.strategy_file_name = None
            self.strategy_name = None
            self.strategy_config = None
            raise
        if all_status_go:
            self.notify("\nEnter \"start\" to start market making.")

    @staticmethod
    def restore_config_legacy(config_map: Dict[str, ConfigVar], config_map_backup: Dict[str, ConfigVar]):
        for key in config_map:
            config_map[key] = config_map_backup[key]

    def reset_application_state(self):
        self.app.change_prompt(prompt=">>> ")
        self.app.input_field.completer = load_completer(self)
        self.placeholder_mode = False
        self.app.hide_input = False



from decimal import Decimal
from typing import Any, Dict, Optional

from pydantic import ConfigDict, SecretStr, field_validator
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = False
DOMAIN = ["tegro"]
EXAMPLE_PAIR = "ZRX-ETH"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0"),
)


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    if isinstance(exchange_info, dict):
        symbol: str = exchange_info.get("symbol", "")
        state: str = exchange_info.get("state", "")
        return True if state == "verified" and symbol.count("_") == 1 else False


def validate_mainnet_exchange(value: str) -> Optional[str]:
    """
    Permissively interpret a string as a boolean
    """
    valid_values = ('base')
    if value.lower() not in valid_values:
        return f"Invalid value, please choose value from {valid_values}"


def validate_testnet_exchange(value: str) -> Optional[str]:
    """
    Permissively interpret a string as a boolean
    """
    valid_values = ('base', 'polygon', 'optimism')
    if value.lower() not in valid_values:
        return f"Invalid value, please choose value from {valid_values}"


def int_val_or_none(string_value: str,
                    on_error_return_none: bool = True,
                    ) -> int:
    try:
        return int(string_value)
    except Exception:
        if on_error_return_none:
            return None
        else:
            return int('0')


def decimal_val_or_none(string_value: str,
                        on_error_return_none: bool = True,
                        ) -> Decimal:
    try:
        return Decimal(string_value)
    except Exception:
        if on_error_return_none:
            return None
        else:
            return Decimal('0')


class TegroConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="tegro", const=True, client_data=None)
    tegro_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Public Wallet Address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    tegro_api_secret: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Private Wallet Address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    chain_name: str = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your preferred chain. (base/ )",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )

    @field_validator("chain_name", mode="before")
    @classmethod
    def validate_exchange(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_mainnet_exchange(v)
            if ret is not None:
                raise ValueError(ret)
        return v
    model_config = ConfigDict(title="tegro")


KEYS = TegroConfigMap.construct()


class TegroTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="tegro_testnet", const=True, client_data=None)
    tegro_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Public Wallet Address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    tegro_api_secret: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Private Wallet Address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    chain_name: str = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your preferred chain. (base/polygon/optimism)",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )

    @field_validator("chain_name", mode="before")
    @classmethod
    def validate_exchange(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_testnet_exchange(v)
            if ret is not None:
                raise ValueError(ret)
        return v
    model_config = ConfigDict(title="tegro_testnet")


OTHER_DOMAINS = ["tegro_testnet"]
OTHER_DOMAINS_PARAMETER = {"tegro_testnet": "tegro_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"tegro_testnet": "BTC-USDT"}
OTHER_DOMAINS_DEFAULT_FEES = {"tegro_testnet": DEFAULT_FEES}
OTHER_DOMAINS_KEYS = {"tegro_testnet": TegroTestnetConfigMap.construct()}



import base64
import hmac
from typing import Any, Dict, List
from urllib.parse import urlencode

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class BitgetPerpetualAuth(AuthBase):
    """
    Auth class required by Bitget Perpetual API
    """
    def __init__(self, api_key: str, secret_key: str, passphrase: str, time_provider: TimeSynchronizer):
        self._api_key: str = api_key
        self._secret_key: str = secret_key
        self._passphrase: str = passphrase
        self._time_provider: TimeSynchronizer = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        headers = {}
        headers["Content-Type"] = "application/json"
        headers["ACCESS-KEY"] = self._api_key
        headers["ACCESS-TIMESTAMP"] = str(int(self._time_provider.time() * 1e3))
        headers["ACCESS-PASSPHRASE"] = self._passphrase
        # headers["locale"] = "en-US"

        path = request.throttler_limit_id
        if request.method is RESTMethod.GET:
            path += "?" + urlencode(request.params)

        payload = str(request.data)
        headers["ACCESS-SIGN"] = self._sign(
            self._pre_hash(headers["ACCESS-TIMESTAMP"], request.method.value, path, payload),
            self._secret_key)
        request.headers.update(headers)
        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. OKX does not use this
        functionality
        """
        return request  # pass-through

    def get_ws_auth_payload(self) -> List[Dict[str, Any]]:
        """
        Generates a dictionary with all required information for the authentication process
        :return: a dictionary of authentication info including the request signature
        """
        timestamp = str(int(self._time_provider.time()))
        signature = self._sign(self._pre_hash(timestamp, "GET", "/user/verify", ""), self._secret_key)
        auth_info = [
            {
                "apiKey": self._api_key,
                "passphrase": self._passphrase,
                "timestamp": timestamp,
                "sign": signature
            }
        ]
        return auth_info

    @staticmethod
    def _sign(message, secret_key):
        mac = hmac.new(bytes(secret_key, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d).decode().strip()

    @staticmethod
    def _pre_hash(timestamp: str, method: str, request_path: str, body: str):
        if body in ["None", "null"]:
            body = ""
        return str(timestamp) + method.upper() + request_path + body



import logging
from decimal import Decimal
from functools import lru_cache
from typing import Dict, List, Optional, Set

from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import ReadOnlyClientConfigAdapter, get_connector_class
from hummingbot.client.config.security import Security
from hummingbot.client.settings import AllConnectorSettings, GatewayConnectionSetting, gateway_connector_trading_pairs
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.utils.gateway_config_utils import flatten
from hummingbot.core.utils.market_price import get_last_price


class UserBalances:
    __instance = None

    @staticmethod
    def connect_market(exchange, client_config_map: ClientConfigMap, **api_details):
        connector = None
        conn_setting = AllConnectorSettings.get_connector_settings()[exchange]
        if api_details or conn_setting.uses_gateway_generic_connector():
            connector_class = get_connector_class(exchange)
            read_only_client_config = ReadOnlyClientConfigAdapter.lock_config(client_config_map)
            init_params = conn_setting.conn_init_parameters(
                trading_pairs=gateway_connector_trading_pairs(conn_setting.name),
                api_keys=api_details,
                client_config_map=read_only_client_config,
            )

            # collect trading pairs from the gateway connector settings
            trading_pairs: List[str] = gateway_connector_trading_pairs(conn_setting.name)

            # collect unique trading pairs that are for balance reporting only
            if conn_setting.uses_gateway_generic_connector():
                config: Optional[Dict[str, str]] = GatewayConnectionSetting.get_connector_spec_from_market_name(conn_setting.name)
                if config is not None:
                    existing_pairs = set(flatten([x.split("-") for x in trading_pairs]))

                    other_tokens: Set[str] = set(config.get("tokens", "").split(","))
                    other_tokens.discard("")
                    tokens: List[str] = [t for t in other_tokens if t not in existing_pairs]
                    if tokens != [""]:
                        trading_pairs.append("-".join(tokens))

            connector = connector_class(**init_params)
        return connector

    # return error message if the _update_balances fails
    @staticmethod
    async def _update_balances(market) -> Optional[str]:
        try:
            await market._update_balances()
        except Exception as e:
            logging.getLogger().debug(f"Failed to update balances for {market}", exc_info=True)
            return str(e)
        return None

    @staticmethod
    def instance():
        if UserBalances.__instance is None:
            UserBalances()
        return UserBalances.__instance

    @staticmethod
    @lru_cache(maxsize=10)
    def is_gateway_market(exchange_name: str) -> bool:
        return (
            exchange_name in sorted(
                AllConnectorSettings.get_gateway_amm_connector_names()
            )
        )

    def __init__(self):
        if UserBalances.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            UserBalances.__instance = self
        self._markets = {}

    async def add_exchange(self, exchange, client_config_map: ClientConfigMap, **api_details) -> Optional[str]:
        self._markets.pop(exchange, None)
        is_gateway_market = self.is_gateway_market(exchange)
        if not is_gateway_market:
            market = UserBalances.connect_market(exchange, client_config_map, **api_details)
            if not market:
                return "API keys have not been added."
            err_msg = await UserBalances._update_balances(market)
            if err_msg is None:
                self._markets[exchange] = market
            return err_msg

    def all_balances(self, exchange) -> Dict[str, Decimal]:
        if exchange not in self._markets:
            return {}
        return self._markets[exchange].get_all_balances()

    async def update_exchange_balance(self, exchange_name: str, client_config_map: ClientConfigMap) -> Optional[str]:
        is_gateway_market = self.is_gateway_market(exchange_name)
        if is_gateway_market and exchange_name in self._markets:
            # we want to refresh gateway connectors always, since the applicable tokens change over time.
            # doing this will reinitialize and fetch balances for active trading pair
            del self._markets[exchange_name]
        if exchange_name in self._markets:
            return await self._update_balances(self._markets[exchange_name])
        else:
            await Security.wait_til_decryption_done()
            api_keys = Security.api_keys(exchange_name) if not is_gateway_market else {}
            return await self.add_exchange(exchange_name, client_config_map, **api_keys)

    # returns error message for each exchange
    async def update_exchanges(
        self,
        client_config_map: ClientConfigMap,
        reconnect: bool = False,
        exchanges: Optional[List[str]] = None
    ) -> Dict[str, Optional[str]]:
        exchanges = exchanges or []
        tasks = []
        # Update user balances
        if len(exchanges) == 0:
            exchanges = [cs.name for cs in AllConnectorSettings.get_connector_settings().values()]
        exchanges: List[str] = [
            cs.name
            for cs in AllConnectorSettings.get_connector_settings().values()
            if not cs.use_ethereum_wallet
            and cs.name in exchanges
            and not cs.name.endswith("paper_trade")
        ]

        if reconnect:
            self._markets.clear()
        for exchange in exchanges:
            tasks.append(self.update_exchange_balance(exchange, client_config_map))
        results = await safe_gather(*tasks)
        return {ex: err_msg for ex, err_msg in zip(exchanges, results)}

    # returns only for non-gateway connectors since balance command no longer reports gateway connector balances
    async def all_balances_all_exchanges(self, client_config_map: ClientConfigMap) -> Dict[str, Dict[str, Decimal]]:
        await self.update_exchanges(client_config_map)
        return {k: v.get_all_balances() for k, v in sorted(self._markets.items(), key=lambda x: x[0]) if not self.is_gateway_market(k)}

    # returns only for non-gateway connectors since balance command no longer reports gateway connector balances
    def all_available_balances_all_exchanges(self) -> Dict[str, Dict[str, Decimal]]:
        return {k: v.available_balances for k, v in sorted(self._markets.items(), key=lambda x: x[0]) if not self.is_gateway_market(k)}

    async def balances(self, exchange, client_config_map: ClientConfigMap, *symbols) -> Dict[str, Decimal]:
        if await self.update_exchange_balance(exchange, client_config_map) is None:
            results = {}
            for token, bal in self.all_balances(exchange).items():
                matches = [s for s in symbols if s.lower() == token.lower()]
                if matches:
                    results[matches[0]] = bal
            return results

    @staticmethod
    def validate_ethereum_wallet() -> Optional[str]:
        return "Connector deprecated."

    @staticmethod
    async def base_amount_ratio(exchange, trading_pair, balances) -> Optional[Decimal]:
        try:
            base, quote = trading_pair.split("-")
            base_amount = balances.get(base, 0)
            quote_amount = balances.get(quote, 0)
            price = await get_last_price(exchange, trading_pair)
            total_value = base_amount + (quote_amount / price)
            return None if total_value <= 0 else base_amount / total_value
        except Exception:
            return None
